import os.path
from typing import IO, List, Optional, Any, Callable, Dict, Set
from enum import Enum
from collections import deque

from ..util.encoding import str_to_bytes

from .lexing import Lexer, Token, TokenType, CharacterType, STRING_ESCAPE


class SourceMetadata:

    def __init__(self, path: bytes):
        self.path = path


class Source:

    def __init__(
                self,
                stream: IO,
                metadata: SourceMetadata
            ):
        self.stream = stream
        self.metadata = metadata


class PhpException(Exception):
    pass


class ParsingException(PhpException):
    pass


class EvaluationException(PhpException):
    pass


class ImplementationException(PhpException):
    pass


class PhpStateType:
    pass


def make_strings_binary(value: Any) -> Any:
    if isinstance(value, str):
        return str_to_bytes(value)
    return value


class PhpType(Enum):
    STRING = bytes,
    INTEGER = int,
    ARRAY = List,
    NULL = type(None)

    def is_valid_value(self, value: Any) -> bool:
        return value is self.value or isinstance(value, self.value)

    def validate(self, value: Any) -> None:
        if not self.is_valid_value(value):
            raise EvaluationException(
                    'Value ' + repr(value) +
                    f' is not valid for type {self.name} ({self.value})'
                )

    @classmethod
    def for_python_value(cls, value: Any):
        value = make_strings_binary(value)
        for type in cls:
            if type.is_valid_value(value):
                return type
        raise ImplementationException(
                'Python value does not have a corresponding PHP type: '
                + repr(value)
            )


class PhpValue:

    def __init__(
                self,
                type: PhpType,
                value: Any
            ):
        type.validate(value)
        self.type = type
        self.value = value

    @classmethod
    def for_python_value(cls, value: Any):
        value = make_strings_binary(value)
        type = PhpType.for_python_value(value)
        return cls(
                type,
                value
            )


PHP_NULL = PhpValue(PhpType.NULL, None)
PHP_VOID = PHP_NULL  # TODO: Differentiate null and void


class Evaluable:

    def evaluate(self, state: PhpStateType) -> PhpValue:
        raise ImplementationException()


class PhpName:

    def __init__(
                self,
                components: List[bytes],
                base=None
            ):
        self.components = components
        self.base = base


class PhpEntity:

    def __init__(self):
        self.comments = []

    def attach_comment(self, comment: bytes) -> None:
        self.comments.append(comment)

    def attach_comments(self, comments: List[bytes]) -> None:
        self.comments.extend(comments)


class PhpIdentifiedEntity(PhpEntity):

    def __init__(self, name: bytes):
        super().__init__()
        self.name = name

    def __str__(self):
        return type(self).__name__ + ':' + self.name


class PhpVariable(PhpIdentifiedEntity):

    def __init__(
                self,
                name: bytes,
                value: PhpValue,
            ):
        super().__init__(name)
        self.value = value

    def assign(self, value: PhpValue):
        self.value = value


class PhpInstruction(PhpEntity, Evaluable):

    def __init__(self):
        pass

    def evaluate(self, state: PhpStateType) -> Any:
        pass


class PhpInstructionGroup(PhpEntity, Evaluable):

    def __init__(
                self,
                instructions: List[PhpInstruction] = None):
        self.instructions = instructions if instructions is not None else []

    def add_instruction(self, instruction: PhpInstruction) -> None:
        self.instructions.append(instruction)

    def evaluate(self, state: PhpStateType) -> PhpValue:
        for instruction in self.instructions:
            instruction.evaluate(state)
        return state.return_value


class PhpArgument(PhpIdentifiedEntity):
    pass


class PhpFunction(PhpEntity, Evaluable):

    def __init__(
                self,
                instruction_group: PhpInstructionGroup = None,
                arguments: List[PhpArgument] = None
            ):
        super().__init__()
        self.instruction_group = instruction_group \
            if instruction_group is not None \
            else PhpInstructionGroup()
        self.arguments = arguments if arguments is not None else []

    def evaluate(self, state: PhpStateType):
        # TODO: Pass arguments
        return self.instruction_group.evaluate(state)


class PhpVisibility(bytes, Enum):
    PRIVATE = b'private'
    PROTECTED = b'protected'
    PUBLIC = b'public'

    @classmethod
    def for_token_type(cls, token_type: TokenType):
        if token_type is TokenType.PRIVATE:
            return cls.PRIVATE
        elif token_type is TokenType.PROTECTED:
            return cls.PROTECTED
        elif token_type is TokenType.PUBLIC:
            return cls.PUBLIC
        else:
            return None


class PhpModifier(bytes, Enum):
    ABSTRACT = b'abstract'
    STATIC = b'static'
    FINAL = b'final'
    READONLY = b'readonly'

    @classmethod
    def for_token_type(cls, token_type: TokenType):
        if token_type is TokenType.ABSTRACT:
            return cls.ABSTRACT
        elif token_type is TokenType.STATIC:
            return cls.STATIC
        elif token_type is TokenType.FINAL:
            return cls.FINAL
        elif token_type is TokenType.READONLY:
            return cls.READONLY
        else:
            return None


class PhpModifierGroup:

    def __init__(
                self,
                visibility: PhpVisibility = PhpVisibility.PUBLIC,
                modifiers: Optional[Set[PhpModifier]] = None):
        self.visibility = visibility
        self.modifiers = modifiers if modifiers is not None else set()


class PhpClassMember(PhpEntity, Evaluable):

    def __init__(
                self,
                name: bytes,
                modifier_group: Optional[PhpModifierGroup] = None
            ):
        self.name = name
        self.modifier_group = modifier_group if modifier_group is not None \
            else PhpModifierGroup()


class PhpProperty(PhpClassMember):

    def evaluate(self, state: PhpStateType) -> PhpValue:
        # TODO: Implement this
        return PHP_NULL


class PhpMethod(PhpClassMember):

    def __init__(
                self,
                name: bytes,
                function: PhpFunction,
                modifier_group: Optional[PhpModifierGroup] = None
            ):
        super().__init__(name, modifier_group)
        self.function = function

    def evaluate(self, state: PhpStateType) -> PhpValue:
        # TODO: Implement this
        return PHP_NULL


class PhpClassConstant(PhpEntity, Evaluable):

    def __init__(
                self,
                class_name: bytes,
                constant_name: bytes
            ):
        self.class_name = class_name
        self.constant_name = constant_name

    def evaluate(self, state: PhpStateType) -> PhpValue:
        cls = state.get_class(self.class_name)
        if cls is None:
            raise EvaluationException(
                    f'Class {self.class_name} does not exist'
                )
        value = cls.get_constant(self.constant_name)
        return PhpValue.for_python_value(value)


class PhpClass(PhpIdentifiedEntity):

    def __init__(
                self,
                name: bytes,
                modifier_group: PhpModifierGroup
            ):
        super().__init__(name)
        self.modifier_group = modifier_group
        self.properties = {}
        self.methods = {}
        self.constants = {}

    def add_property(self, property: PhpProperty) -> None:
        self.properties[property.name] = property

    def add_method(self, method: PhpMethod) -> None:
        self.methods[method.name] = method

    def get_method(self, name: bytes) -> Optional[PhpMethod]:
        try:
            return self.methods[name]
        except KeyError:
            return None

    def add_constant(self, constant: PhpClassConstant) -> None:
        self.constants[constant.constant_name] = constant

    def get_constant(self, name: bytes) -> Optional[PhpClassConstant]:
        try:
            return self.constants[name]
        except KeyError:
            return None


class PhpDefinitions:

    def __init__(
                self,
                base_functions: Dict[bytes, Callable] = None,
                base_classes: Dict[bytes, PhpClass] = None
            ):
        self.functions = base_functions.copy()
        self.classes = base_classes.copy()

    def define_function(self, name: bytes, callable: Callable) -> None:
        self.functions[name] = callable

    def get_function(self, name: bytes) -> Optional[Callable]:
        try:
            return self.functions[name]
        except KeyError:
            return None

    def define_class(self, definition: PhpClass) -> None:
        self.classes[definition.name] = definition

    def get_class(self, name: bytes) -> Optional[PhpClass]:
        try:
            return self.classes[name]
        except KeyError:
            return None


class PhpScope:

    def __init__(self):
        self.variables = {}

    def get_variable(self, name: bytes) -> PhpVariable:
        try:
            return self.variables[name]
        except KeyError:
            variable = PhpVariable(
                    name=name,
                    value=PHP_NULL
                )
            self.variables[name] = variable
            return variable


class PhpEvaluationOptions:

    def __init__(
                self,
                allow_includes: bool = True
            ):
        self.allow_includes = allow_includes


class PhpState(PhpStateType):

    def __init__(
                self,
                definitions: PhpDefinitions,
                options: Optional[PhpEvaluationOptions] = None
            ):
        self.definitions = definitions
        self.global_scope = PhpScope()
        self.scope = self.global_scope
        self.constants = {}
        self.output = []
        self.return_value = None
        self.options = options if options is not None \
            else PhpEvaluationOptions()

    def define_constant(self, name: bytes, value: Any) -> None:
        if name in self.constants:
            raise EvaluationException(f'Constant {name} is already defined')
        self.constants[name] = value

    def get_constant(
                self,
                name: bytes,
                default_to_name: bool = True
            ) -> PhpValue:
        try:
            return self.constants[name]
        except KeyError:
            if default_to_name:
                return PhpValue(PhpType.STRING, name)
            return PHP_NULL

    def get_constant_value(
                self,
                name: bytes,
                default_to_name: bool = True
            ) -> PhpValue:
        return self.get_constant(name, default_to_name).value

    def get_variable(self, name: bytes) -> PhpVariable:
        return self.scope.get_variable(name)

    def get_variable_value(self, name: bytes) -> Any:
        return self.get_variable(name).value.value

    def write_output(self, output: bytes) -> None:
        self.output.append(output)


def php_define(state: PhpState, constant: PhpValue, value: PhpValue) -> None:
    state.define_constant(constant.value, value)


def php_defined(state: PhpState, constant: PhpValue) -> PhpValue:
    return PhpValue.for_python_value(constant.value in state.constants)


def php_dirname(state: PhpState, path: PhpValue) -> PhpValue:
    return PhpValue.for_python_value(
            os.fsencode(os.path.dirname(path.value))
        )


BASE_FUNCTIONS = {
        b'define': php_define,
        b'defined': php_defined,
        b'dirname': php_dirname
    }
BASE_CLASSES = {
    }


def initialize_php_definitions() -> PhpDefinitions:
    return PhpDefinitions(
            base_functions=BASE_FUNCTIONS,
            base_classes=BASE_CLASSES
        )


class PhpLiteral(PhpValue, PhpEntity, Evaluable):

    def evaluate(self, state: PhpState) -> PhpValue:
        return self


OPERATOR_MAP = dict()


class PhpOperator(PhpEntity):
    pass


class PhpUnaryOperator(PhpOperator):

    def __init__(
                self,
                operator: bytes,
                callable: Callable[[Any], Any]
            ):
        self.operator = operator
        self.callable = callable

    def apply(self, value: PhpValue) -> Any:
        return self.callable(value)


def _register_unary_operator(
            operator: bytes,
            callable: Callable[[Any], Any]
        ) -> None:
    instance = PhpUnaryOperator(operator, callable)
    OPERATOR_MAP[operator] = instance


_register_unary_operator(
        b'!',
        lambda value: PhpValue(value.type, not value.value)
    )


class PhpBinaryOperator(PhpOperator):

    def __init__(
                self,
                operator: bytes,
                callable: Callable[[Any, Any], Any]
            ):
        super().__init__()
        self.operator = operator
        self.callable = callable

    def apply(self, left: Any, right: Any) -> Any:
        return self.callable(left, right)


def _register_binary_operator(
            operator: bytes,
            callable: Callable[[Any, Any], Any]
        ):
    operator_instance = PhpBinaryOperator(operator, callable)
    OPERATOR_MAP[operator] = operator_instance


_register_binary_operator(
        b'.',
        lambda left, right: PhpValue(left.type, left.value + right.value)
    )
_register_binary_operator(
        b'===',
        lambda left, right: left.type is right.type and
        left.value == right.value
    )
_register_binary_operator(
        b'!==',
        lambda left, right: left.type is not right.type or
        left.value != right.value
    )
_register_binary_operator(
        b'==',
        lambda left, right: left.value == right.value
    )
_register_binary_operator(
        b'!=',
        lambda left, right: left.value != right.value
    )
_register_binary_operator(
        b'=',
        lambda left, right: left.assign(right)
    )
_register_binary_operator(
        b'>=',
        lambda left, right: left.value >= right.value
    )
_register_binary_operator(
        b'&&',
        lambda left, right: left.value and right.value
    )
_register_binary_operator(
        b'||',
        lambda left, right: left.value or right.value
    )


class PhpOutput(PhpInstruction):

    def __init__(self, content: bytes):
        super().__init__()
        self.content = content

    def evaluate(self, state: PhpState) -> PhpValue:
        state.write_output(self.content)
        return PHP_VOID


class PhpExpression(PhpInstruction):

    def __init__(self):
        self.components = []

    def add_component(self, component: PhpEntity) -> None:
        self.components.append(component)

    def evaluate(self, state: PhpState) -> PhpValue:
        operator = None
        value = None
        for component in self.components:
            if isinstance(component, Evaluable):
                new_value = component.evaluate(state)
                if operator is None:
                    if value is not None:
                        raise EvaluationException(
                                'Unexpected adjacent expressions'
                            )
                    value = new_value
                elif isinstance(operator, PhpUnaryOperator):
                    value = operator.apply(new_value)
                elif isinstance(operator, PhpBinaryOperator):
                    value = operator.apply(value, new_value)
                else:
                    raise EvaluationException(
                            'Unexpected operator type'
                        )
            elif isinstance(component, PhpOperator):
                if operator is not None:
                    raise EvaluationException('Unexpected adjacent operators')
                operator = component
            else:
                raise ImplementationException('Not yet implemented')
        if value is None:
            return PHP_NULL
        return value


class PhpDeclaration(PhpIdentifiedEntity):

    def __init__(self):
        pass


class PhpIdentifier(PhpEntity):

    def __init__(self, name: bytes, parent_name: Optional[bytes] = None):
        super().__init__()
        self.name = name
        self.parent_name = parent_name


class PhpVariableReference(PhpIdentifiedEntity, Evaluable):

    def evaluate(self, state: PhpState) -> PhpVariable:
        return state.scope.get_variable(self.name)


class PhpConstant(PhpIdentifiedEntity, Evaluable):

    def evaluate(self, state: PhpState) -> Any:
        try:
            return state.constants[self.name]
        except KeyError:
            # Constants are treated as strings if undefined
            return PhpValue.for_python_value(self.name)


class PhpMagicConstant(PhpEntity, Evaluable):

    def __init__(
                self,
                token_type: TokenType,
                source_metadata: SourceMetadata
            ):
        self.token_type = token_type
        self.source_metadata = source_metadata

    def evaluate(self, state: PhpState) -> Any:
        if self.token_type == TokenType.DIR:
            return PhpValue.for_python_value(
                    os.fsdecode(os.path.dirname(self.source_metadata.path))
                )
        elif self.token_type == TokenType.FILE:
            return PhpValue.for_python_value(
                    os.fsdecode(self.source_metadata.path)
                )
        else:
            raise EvaluationException('Unsupported magic constant')


class PhpInclude(PhpInstruction, Evaluable):

    def __init__(
                self,
                path: PhpExpression,
                required: bool = False,
                once: bool = False
            ):
        self.path = path
        self.required = required
        self.once = once

    def evaluate_path(self, state: PhpState) -> bytes:
        path = self.path.evaluate(state)
        if isinstance(path, bytes):
            return os.fsencode(path)
        raise EvaluationException(
                'Included path is not a string, received: {repr(path)}'
            )

    def evaluate(self, state: PhpState) -> Any:
        if not state.options.allow_includes:
            return None
        path = self.evaluate_path(state)
        context = parse_php_file(path)
        return context.evaluate(state=state)


class PhpCallable():

    def get_callable(self, state: PhpState) -> Callable:
        raise EvaluationException('get_callable is not implemented')


class PhpFunctionReference(PhpIdentifiedEntity, PhpCallable):

    def get_callable(self, state: PhpState) -> Callable:
        callable = state.definitions.get_function(self.name)
        if callable is None:
            raise EvaluationException(f'Function {self.name} is not defined')
        return callable


class PhpStaticMethodReference(PhpCallable):

    def __init__(self, class_name: bytes, method_name: bytes):
        self.class_name = class_name
        self.method_name = method_name

    def get_callable(self, state: PhpState) -> Callable:
        class_definition = state.definitions.get_class(self.class_name)
        callable = class_definition.get_method(self.method_name)
        if callable is None:
            raise EvaluationException(
                    f'Method {self.name} does not exist on class '
                    f'{self.class_name}'
                )
        return callable


class PhpMethodReference(PhpCallable):

    def __init__(
                self,
                method_name: bytes
            ):
        self.method_name = method_name


class PhpCallableInvocation(PhpEntity, Evaluable):

    def __init__(
                self,
                callable: PhpCallable,
                arguments: List[PhpExpression]
            ):
        self.callable = callable
        self.arguments = arguments

    def evaluate(self, state: PhpState) -> Any:
        arguments = [argument.evaluate(state) for argument in self.arguments]
        callable = self.callable.get_callable(state)
        return callable(state, *arguments)

    def __repr__(self) -> str:
        return repr(self.callable) + '(' + repr(self.arguments) + ')'


class PhpPropertyReference(PhpEntity, Evaluable):

    def __init__(
                self,
                property_name: bytes
            ):
        self.property_name = property_name


class PhpCondition(PhpEntity, Evaluable):

    def __init__(
                self,
                instruction_group: PhpInstructionGroup,
                expression: Optional[PhpExpression] = None
            ):
        self.expression = expression
        self.instruction_group = instruction_group

    def should_execute(self, state: PhpState) -> bool:
        if self.expression is None:
            return True
        return bool(self.expression.evaluate(state))

    def evaluate(self, state: PhpState) -> bool:
        if self.should_execute(state):
            self.instruction_group.evaluate(state)
            return True
        return False


class PhpConditional(Evaluable):

    def __init__(
                self,
                conditions: List[PhpCondition]
            ):
        self.conditions = conditions

    def evaluate(self, state: PhpState) -> None:
        for condition in self.conditions:
            if condition.evaluate(state):
                break


class PhpReturn(PhpInstruction):

    def __init__(self, expression: Optional[PhpExpression] = None):
        self.expression = expression

    def evaluate_return_value(self, state: PhpState) -> Any:
        if self.expression is None:
            return None
        return self.expression.evaluate(state)

    def evaluate(self, state: PhpState) -> Any:
        state.return_value = self.evaluate_return_value(state)


class PhpInstantiation(PhpInstruction):

    def __init__(
                self,
                class_name: PhpName,
                arguments: List[PhpExpression]
            ):
        self.class_name = class_name
        self.arguments = arguments

    def evaluate(self, state: PhpState) -> PhpValue:
        raise ImplementationException()


class PhpForeach(PhpInstruction):

    def __init__(
                self,
                expression: PhpExpression,
                instruction_group: PhpInstructionGroup,
                value_name: bytes,
                key_name: Optional[bytes]
            ):
        self.expression = expression
        self.instruction_group = instruction_group
        self.value_name = value_name
        self.key_name = key_name

    def evaluate(self, state: PhpState) -> PhpValue:
        return PHP_VOID


class PhpContext:

    def __init__(self):
        self.instructions = []
        self.definitions = initialize_php_definitions()
        self.state = PhpState(self.definitions)

    def get_includes(self) -> List[PhpInclude]:
        includes = []
        for instruction in self.instructions:
            if isinstance(instruction, PhpInclude):
                includes.append(instruction)
        return includes

    def evaluate(
                self,
                state: PhpState = None,
                options: PhpEvaluationOptions = None
            ) -> PhpState:
        if state is None:
            state = PhpState(
                    definitions=self.definitions
                )
        state.options = options
        for instruction in self.instructions:
            instruction.evaluate(state)
        return state


COMMENT_TOKEN_TYPES = {
        TokenType.COMMENT,
        TokenType.DOC_COMMENT
    }
INCLUDE_TOKEN_TYPES = {
        TokenType.INCLUDE,
        TokenType.INCLUDE_ONCE,
        TokenType.REQUIRE,
        TokenType.REQUIRE_ONCE
    }
REQUIRE_TOKEN_TYPES = {
        TokenType.REQUIRE,
        TokenType.REQUIRE_ONCE
    }
INCLUDE_ONCE_TOKEN_TYPES = {
        TokenType.INCLUDE_ONCE,
        TokenType.REQUIRE_ONCE
    }
MAGIC_CONSTANT_TOKEN_TYPES = {
        TokenType.LINE,
        TokenType.FILE,
        TokenType.DIR,
        TokenType.CLASS_C,
        TokenType.TRAIT_C,
        TokenType.METHOD_C,
        TokenType.FUNC_C,
        TokenType.NS_C
    }
SECONDARY_CONDITION_TOKEN_TYPES = {
        TokenType.ELSEIF,
        TokenType.ELSE
    }


class TokenStream:

    def __init__(self, lexer: Lexer):
        self.lexer = lexer
        self.pending_comments = []
        self.pending_tokens = deque()

    def consume_preview(self) -> None:
        try:
            return self.pending_tokens.popleft()
        except IndexError:
            raise ParsingException('No preview tokens present')

    def accept_base_token(
                self,
                include_pending: bool = True
            ) -> Optional[Token]:
        if include_pending:
            try:
                return self.pending_tokens.popleft()
            except IndexError:
                pass  # No pending tokens
        return self.lexer.get_next_token()

    def accept_token(self, include_pending: bool = True) -> Optional[Token]:
        while (token := self.accept_base_token()) is not None:
            if token.type is TokenType.WHITESPACE:
                continue
            elif token.type in COMMENT_TOKEN_TYPES:
                self.pending_comments.append(token.value)
            else:
                return token
        return None

    def require_token(self) -> Token:
        token = self.accept_token()
        if token is None:
            raise ParsingException('Token expected')
        return token

    def require_token_of_type(self, type: TokenType) -> Token:
        token = self.require_token()
        if token.type is not type:
            raise ParsingException(f'Expected token of type {type}')
        return token

    def require_token_of_types(self, types: Set[TokenType]) -> Token:
        token = self.require_token()
        if token.type not in types:
            types_string = ', '.join(types)
            raise ParsingException(
                    'Expected token of one of the following types: '
                    f'{types_string}'
                )
        return token

    def require_semicolon(self) -> None:
        token = self.require_token()
        if token.is_semicolon():
            return
        raise ParsingException('Expected semicolon')

    def require_character(self, type: CharacterType) -> None:
        token = self.require_token()
        if token.type == TokenType.CHARACTER \
                and token.value == type:
            return
        raise ParsingException(f'Expected character: {type}')

    def require_equals(self) -> None:
        self.require_character(CharacterType.EQUALS)

    def take_comments(self) -> List[bytes]:
        comments = self.pending_comments
        self.pending_comments = []
        return comments

    def push_token(self, token: Token) -> None:
        self.pending_tokens.append(token)

    def preview_token(self) -> Optional[Token]:
        next = self.accept_token(False)
        if next is not None:
            self.push_token(next)
        return next

    def require_preview_token(self) -> Token:
        token = self.preview_token()
        if token is None:
            raise ParsingException('Unexpected end of token stream')
        return token


class TagStateChanged(Exception):

    def __init__(self, state: bool):
        self.state = state


expression_level = 0


class Parser:

    def __init__(self, source: Source):
        self.source = source
        self.lexer = Lexer(source.stream)
        self.token_stream = TokenStream(self.lexer)

    def parse_instruction(
                self,
                token_stream: TokenStream
            ) -> Optional[PhpInstruction]:
        return None

    def parse_string(self, token: Token) -> PhpLiteral:
        if token.type != TokenType.CONSTANT_ENCAPSED_STRING:
            raise ParsingException('Token is not a valid string')
        value = token.value[1:-1].replace(STRING_ESCAPE, b'')
        return PhpLiteral(PhpType.STRING, value)

    def parse_integer(self, token: Token) -> PhpLiteral:
        if token.type != TokenType.LNUMBER:
            raise ParsingException('Token is not a valid integer literal')
        value = int(token.value)
        return PhpLiteral(PhpType.INTEGER, value)

    def parse_magic_constant(self, token: Token) -> PhpMagicConstant:
        if token.type not in MAGIC_CONSTANT_TOKEN_TYPES:
            raise ParsingException('Token is not a magic constant')
        return PhpMagicConstant(token.type, self.source.metadata)

    def parse_constant(self, identifier: PhpIdentifier) -> PhpConstant:
        if identifier.parent_name is not None:
            return PhpClassConstant(identifier.parent_name, identifier.name)
        return PhpConstant(identifier.name)

    def parse_operator(self, token: Token) -> PhpBinaryOperator:
        try:
            return OPERATOR_MAP[token.value]
        except KeyError:
            raise ParsingException(f'Unrecognized operator: {token}')

    def parse_identifier(
                self,
                token: Token,
                token_stream: TokenStream
            ) -> PhpIdentifier:
        if token.type is not TokenType.STRING:
            raise ParsingException(
                    'First token of identifier is not the correct type'
                )
        name = token.value
        parent_name = None
        preview = token_stream.preview_token()
        if preview.type is TokenType.PAAMAYIM_NEKUDOTAYIM:
            parent_name = name
            token_stream.consume_preview()
            preview = token_stream.preview_token()
            name = token_stream.require_token_of_types(
                    {TokenType.STRING, TokenType.VARIABLE}
                ).value
        return PhpIdentifier(name, parent_name)

    def parse_array(
                self,
                token: Token,
                token_stream: TokenStream
            ) -> PhpLiteral:
        if token.type is not TokenType.ARRAY:
            raise ParsingException(f'Expected array, received: {token}')
        items = self.parse_argument_list(token_stream)
        return PhpLiteral(PhpType.ARRAY, items)

    def parse_name(
                self,
                token_stream: TokenStream
            ) -> PhpName:
        components = []
        base = None  # TODO: Default to current namespace
        first = True
        accept_component = True
        while True:
            token = token_stream.require_preview_token()
            if first:
                first = False
                if token.type is TokenType.NS_SEPARATOR:
                    token_stream.consume_preview()
                    base = None
                    continue
            if token.type is TokenType.STRING:
                if not accept_component:
                    raise ParsingException('Missing namespace separator')
                components.append(token.value)
            elif token.type is TokenType.NS_SEPARATOR:
                if accept_component:
                    raise ParsingException('Missing name component')
            else:
                break
            token_stream.consume_preview()
            accept_component = not accept_component
        if len(components) == 0:
            raise ParsingException('Name is empty')
        return PhpName(
                components,
                base
            )

    def parse_instantiation(
                self,
                token: Token,
                token_stream: TokenStream
            ) -> PhpInstantiation:
        if token.type is not TokenType.NEW:
            raise ParsingException('Expected new, received: {token}')
        class_name = self.parse_name(token_stream)
        arguments = self.parse_argument_list(token_stream)
        return PhpInstantiation(
                class_name,
                arguments
            )

    def parse_object_access(
                self,
                token_stream: TokenStream
            ):
        token = token_stream.requir_token()
        if token.type is TokenType.STRING:
            preview = token_stream.require_preview_token()
            if preview.is_opening_parenthesis():
                arguments = self.parse_argument_list(token_stream)
                return PhpCallableInvocation(
                        PhpMethod(token.value),
                        arguments
                    )
            else:
                pass

    def parse_expression_component(
                self,
                token: Token,
                token_stream: TokenStream
            ) -> PhpEntity:
        # TODO: Parse more expression components
        if token.type is TokenType.CONSTANT_ENCAPSED_STRING:
            return self.parse_string(token)
        elif token.type is TokenType.LNUMBER:
            return self.parse_integer(token)
        elif token.type in MAGIC_CONSTANT_TOKEN_TYPES:
            return self.parse_magic_constant(token)
        elif token.type is TokenType.STRING:
            if token.value.lower() == 'null':
                return PHP_NULL
            identifier = self.parse_identifier(token, token_stream)
            preview = token_stream.preview_token()
            if preview is not None:
                if preview.is_character(CharacterType.OPEN_PARENTHESIS):
                    return self.parse_invocation(
                            identifier,
                            token_stream
                        )
            return self.parse_constant(identifier)
        elif token.type is TokenType.VARIABLE:
            return PhpVariableReference(
                    self.parse_variable_name(token)
                )
        elif token.type is TokenType.ARRAY:
            return self.parse_array(token, token_stream)
        elif token.type is TokenType.NEW:
            return self.parse_instantiation(token, token_stream)
        elif token.type is TokenType.NS_SEPARATOR:
            # TODO: Apply root namespace to subsequent element
            next = token_stream.require_token()
            return self.parse_expression_component(next, token_stream)
        elif token.type in INCLUDE_TOKEN_TYPES:
            return self.parse_include(token, token_stream)
        elif token.type is TokenType.OBJECT_OPERATOR:
            return self.parse_object_access(token_stream)
        elif token.value in OPERATOR_MAP:
            return self.parse_operator(token)
        else:
            raise ParsingException(
                    f'Unrecognized token in expression: {token}'
                )

    def parse_expression(
                self,
                token_stream: TokenStream,
                nested: bool = False,
                terminal_tokens: Set[TokenType] = None
            ) -> PhpExpression:
        global expression_level
        expression_level += 1
        expression = PhpExpression()
        while True:
            token = token_stream.accept_token()
            if token is None or token.is_semicolon():
                if nested:
                    token_stream.push_token(token)
                break
            if token.is_character({
                        CharacterType.COMMA,
                        CharacterType.CLOSE_PARENTHESIS,
                        CharacterType.CLOSE_BRACE
                    }) or (
                        terminal_tokens is not None
                        and token.type in terminal_tokens
                    ):
                token_stream.push_token(token)
                break
            if token.is_character(CharacterType.OPEN_PARENTHESIS):
                component = self.parse_expression(token_stream, True)
                token_stream.require_character(CharacterType.CLOSE_PARENTHESIS)
            else:
                component = self.parse_expression_component(
                            token,
                            token_stream
                        )
            expression.add_component(component)
        expression_level -= 1
        return expression

    def parse_variable_name(self, token: Token) -> bytes:
        return token.value[1:]

    def parse_include(
                self,
                token: Token,
                token_stream: TokenStream,
            ) -> PhpInclude:
        expression = self.parse_expression(token_stream, True)
        return PhpInclude(
                expression,
                required=token in REQUIRE_TOKEN_TYPES,
                once=token in INCLUDE_ONCE_TOKEN_TYPES
            )

    def parse_argument_list(
                self,
                token_stream: TokenStream
            ) -> List:
        token_stream.require_character(CharacterType.OPEN_PARENTHESIS)
        arguments = []
        while True:
            next = token_stream.accept_token()
            if next.is_character(CharacterType.CLOSE_PARENTHESIS):
                break
            if next.is_character(CharacterType.COMMA):
                continue
            token_stream.push_token(next)
            expression = self.parse_expression(token_stream, True)
            arguments.append(expression)
        return arguments

    def parse_invocation(
                self,
                identifier: PhpIdentifier,
                token_stream: TokenStream
            ) -> PhpCallableInvocation:
        arguments = self.parse_argument_list(token_stream)
        if identifier.parent_name is None:
            callable = PhpFunctionReference(identifier.name)
        else:
            callable = PhpStaticMethodReference(
                        identifier.parent_name,
                        identifier.name
                    )
        return PhpCallableInvocation(
                callable,
                arguments
            )

    def parse_return(
                self,
                token_stream: TokenStream
            ) -> PhpReturn:
        token_stream.require_token_of_type(TokenType.RETURN)
        preview = token_stream.preview_token()
        if preview.is_semicolon():
            expression = None
        else:
            expression = self.parse_expression(token_stream, True)
        return PhpReturn(expression)

    def parse_modifiers(
                self,
                token_stream: TokenStream,
                allowed: Set = None
            ) -> PhpModifierGroup:
        group = PhpModifierGroup()
        visibility_specified = False
        while True:
            token = token_stream.preview_token()
            if token is None:
                break
            visibility = PhpVisibility.for_token_type(token.type)
            if visibility is not None:
                if visibility_specified:
                    raise ParsingException(
                            'Multiple visibility modifiers found'
                        )
                if allowed is not None and visibility not in allowed:
                    raise ParsingException(
                            f'Visibility {visibility} is not permitted'
                        )
                group.visibility = visibility
                visibility_specified = True
                token_stream.consume_preview()
                continue
            modifier = PhpModifier.for_token_type(token.type)
            if modifier is not None:
                if allowed is not None and modifier not in allowed:
                    raise ParsingException(
                            f'Modifier {modifier} is not permitted'
                        )
                group.modifiers.add(modifier)
                token_stream.consume_preview()
                continue
            break
        return group

    def parse_property(
                self,
                token: Token,
                token_stream: TokenStream,
                modifier_group: PhpModifierGroup
            ) -> PhpProperty:
        name = self.parse_variable_name(token)
        token_stream.require_semicolon()
        property = PhpProperty(name, modifier_group)
        # TODO: Parse default values
        return property

    def parse_arguments(
                self,
                token_stream: TokenStream
            ) -> List[PhpArgument]:
        arguments = []
        token_stream.require_character(CharacterType.OPEN_PARENTHESIS)
        while (token := token_stream.preview_token()) is not None:
            if token.is_closing_parenthesis():
                break
            if token.type is TokenType.VARIABLE:
                token_stream.consume_preview()
                arguments.append(PhpArgument(token.value))
            token = token_stream.preview_token()
            if token is None or token.is_closing_parenthesis():
                break
            if token.is_comma():
                token_stream.consume_preview()
                continue
            raise ParsingException(
                    f'Unexpected token in function arguments: {token}'
                )
        token_stream.require_character(CharacterType.CLOSE_PARENTHESIS)
        return arguments

    def parse_function(
                self,
                token_stream: TokenStream,
            ) -> PhpFunction:
        arguments = self.parse_arguments(token_stream)
        token_stream.require_character(CharacterType.OPEN_BRACE)
        instruction_group = PhpInstructionGroup()
        in_php_tag = True
        while True:
            preview = token_stream.preview_token()
            if preview.is_character(CharacterType.CLOSE_BRACE):
                token_stream.consume_preview()
                break
            else:
                in_php_tag, instruction = self.parse_any(
                        token_stream,
                        in_php_tag
                    )
                if instruction is None:
                    break
                instruction_group.add_instruction(instruction)
        return PhpFunction(
                instruction_group,
                arguments
            )

    def parse_method(
                self,
                token_stream: TokenStream,
                modifier_group: PhpModifierGroup
            ) -> PhpMethod:
        name = token_stream.require_token_of_type(TokenType.STRING).value
        function = self.parse_function(token_stream)
        return PhpMethod(
                name,
                function,
                modifier_group
            )

    def parse_members(
                self,
                token_stream: TokenStream,
                definition: PhpClass
            ) -> None:
        while True:
            modifier_group = self.parse_modifiers(token_stream)
            token = token_stream.preview_token()
            if token is None:
                break
            if token.type is TokenType.VARIABLE:
                token_stream.consume_preview()
                property = self.parse_property(
                        token,
                        token_stream,
                        modifier_group
                    )
                definition.add_property(property)
            elif token.type is TokenType.FUNCTION:
                token_stream.consume_preview()
                method = self.parse_method(
                        token_stream,
                        modifier_group
                    )
                definition.add_method(method)
            elif token.is_character(CharacterType.CLOSE_BRACE):
                break
            else:
                raise ParsingException(
                        f'Unexpected token in class definition: {token}'
                    )

    def parse_class(
                self,
                token_stream: TokenStream
            ) -> PhpClass:
        modifier_group = self.parse_modifiers(
                token_stream,
                {
                    PhpModifier.ABSTRACT,
                    PhpModifier.FINAL
                }
            )
        token_stream.require_token_of_type(TokenType.CLASS)
        name = token_stream.require_token_of_type(TokenType.STRING).value
        definition = PhpClass(
                name,
                modifier_group
            )
        token_stream.require_character(CharacterType.OPEN_BRACE)
        self.parse_members(token_stream, definition)
        token_stream.require_character(CharacterType.CLOSE_BRACE)
        return definition

    def parse_block(
                self,
                token_stream: TokenStream
            ) -> PhpInstructionGroup:
        instruction_group = PhpInstructionGroup()
        next = token_stream.require_preview_token()
        if next.is_character(CharacterType.OPEN_BRACE):
            token_stream.require_character(CharacterType.OPEN_BRACE)
            while True:
                next = token_stream.require_preview_token()
                if next.is_character(CharacterType.CLOSE_BRACE):
                    break
                instruction = self.parse_statement(token_stream)
                instruction_group.add_instruction(instruction)
            token_stream.require_character(CharacterType.CLOSE_BRACE)
        else:
            instruction = self.parse_statement(token_stream)
            instruction_group.add_instruction(instruction)
        return instruction_group

    def parse_condition(
                self,
                token_stream: TokenStream,
                first: bool = True
            ) -> Optional[PhpCondition]:
        token = token_stream.preview_token()
        if first:
            token_stream.require_token()
            if token.type is not TokenType.IF:
                raise ParsingException(
                        f'Unexpected token in conditional: {token}'
                    )
        else:
            if token is None or \
                    token.type not in SECONDARY_CONDITION_TOKEN_TYPES:
                return None
            token_stream.consume_preview()
        if token.type is TokenType.ELSE:
            expression = None
        else:
            token_stream.require_character(CharacterType.OPEN_PARENTHESIS)
            expression = self.parse_expression(token_stream, True)
            token_stream.require_character(CharacterType.CLOSE_PARENTHESIS)
        instruction_group = self.parse_block(token_stream)
        return PhpCondition(instruction_group, expression)

    def parse_conditional(
                self,
                token_stream: TokenStream
            ) -> PhpConditional:
        first = True
        conditions = []
        while True:
            condition = self.parse_condition(token_stream, first)
            if condition is None:
                if first:
                    raise ParsingException('Malformed condition')
                break
            conditions.append(condition)
            first = False
        return PhpConditional(conditions)

    def parse_foreach(
                self,
                token_stream: TokenStream
            ) -> PhpForeach:
        token_stream.require_token_of_type(TokenType.FOREACH)
        token_stream.require_character(CharacterType.OPEN_PARENTHESIS)
        expression = self.parse_expression(token_stream, True, {TokenType.AS})
        token_stream.require_token_of_type(TokenType.AS)
        first_name = token_stream.require_token_of_type(
                TokenType.VARIABLE
            ).value
        preview = token_stream.require_preview_token()
        if preview.type is TokenType.DOUBLE_ARROW:
            token_stream.consume_preview()
            second_name = token_stream.require_token_of_type(
                    TokenType.VARIABLE
                ).value
        elif preview.is_closing_parenthesis():
            second_name = None
        token_stream.require_character(CharacterType.CLOSE_PARENTHESIS)
        instruction_group = self.parse_block(token_stream)
        if second_name is None:
            value_name = first_name
            key_name = None
        else:
            key_name = first_name
            value_name = second_name
        return PhpForeach(
                expression,
                instruction_group,
                value_name,
                key_name
            )

    def parse_output(
                self,
                token_stream: TokenStream
            ) -> Optional[PhpOutput]:
        content = None
        while (token := token_stream.accept_base_token()) is not None:
            if token.type is TokenType.INLINE_HTML:
                if content is None:
                    content = token.value
                else:
                    content = content + token.value
            elif token.type is TokenType.OPEN_TAG:
                if content is None:
                    raise TagStateChanged(True)
                else:
                    break
            elif token.type is TokenType.CLOSE_TAG:
                raise TagStateChanged(False)
            else:
                raise ParsingException(f'Unexpected token: {token}')
        if content is not None:
            return PhpOutput(content)
        return None

    def parse_statement(
                self,
                token_stream: TokenStream,
                in_php_tag: bool = False
            ) -> Optional[PhpInstruction]:
        preview = token_stream.preview_token()
        if preview is None:
            return None
        elif preview.type is TokenType.CLOSE_TAG:
            raise TagStateChanged(False)
        elif preview.type is TokenType.RETURN:
            return self.parse_return(token_stream)
        elif preview.type is TokenType.CLASS:
            return self.parse_class(token_stream)
        elif preview.type is TokenType.IF:
            return self.parse_conditional(token_stream)
        elif preview.type is TokenType.FOREACH:
            return self.parse_foreach(token_stream)
        else:
            return self.parse_expression(token_stream)

    def parse_any(
                self,
                token_stream: TokenStream,
                in_php_tag: bool = False
            ) -> (bool, Optional[PhpInstruction]):
        try:
            if in_php_tag:
                return (True, self.parse_statement(token_stream))
            else:
                return (False, self.parse_output(token_stream))
        except TagStateChanged as change:
            return self.parse_any(token_stream, change.state)

    def parse(self, context: PhpContext = None) -> PhpContext:
        if context is None:
            context = PhpContext()
        in_php_tag = False
        while True:
            in_php_tag, instruction = self.parse_any(
                    self.token_stream,
                    in_php_tag
                )
            if instruction is None:
                break
            context.instructions.append(instruction)
        return context


def parse_php_file(path: bytes) -> PhpContext:
    try:
        with open(path, 'rb') as stream:
            metadata = SourceMetadata(path)
            source = Source(stream, metadata)
            parser = Parser(source)
            return parser.parse()
    except OSError as error:
        raise ParsingException(
                f'Unable to read file at {path}'
            ) from error



from wordfence.util.pcre import PCRE_DEFAULT_MATCH_LIMIT, \
        PCRE_DEFAULT_MATCH_LIMIT_RECURSION
from wordfence.util.units import byte_length

from ...scanning.matching import MatchEngine
from ..subcommands import SubcommandDefinition, UsageExample
from ..config.typing import ConfigDefinitions
from .reporting import SCAN_REPORT_CONFIG_OPTIONS


config_definitions: ConfigDefinitions = {
    "read-stdin": {
        "description": "Read paths from stdin. If not specified, paths will "
                       "automatically be read from stdin when input is not "
                       "from a TTY.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "file-list-separator": {
        "short_name": "s",
        "description": "Separator used when listing files via stdin. Defaults "
                       "to the null byte.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "AA==",
        "default_type": "base64"
    },
    **SCAN_REPORT_CONFIG_OPTIONS,
    "exclude-signatures": {
        "short_name": "e",
        "description": "Specify rule IDs to exclude from the scan. Can be "
                       "comma-delimited and/or specified multiple times.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ",",
            "value_type": int
        }
    },
    "include-signatures": {
        "short_name": "i",
        "description": "Specify rule IDs to include in the scan. Can be "
                       "comma-delimited and/or specified multiple times.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ",",
            "value_type": int
        }
    },
    "images": {
        "description": "Include image files in the scan.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    },
    "include-all-files": {
        "short_name": "a",
        "description": "Scan all files. By default, only files matching "
                       "certain extensions are scanned",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    },
    "include-files": {
        "short_name": "n",
        "description": "Only scan filenames that are exact matches. Can be "
                       "used multiple times.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ",",
            "accepts_file": True
        }
    },
    "exclude-files": {
        "short_name": "x",
        "description": "Do not scan filenames that are exact matches. Can be "
                       "used multiple times. Denials take precedence over "
                       "allows.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ",",
            "accepts_file": True
        }
    },
    "include-files-pattern": {
        "short_name": "N",
        "description": "Python regex allow pattern. Only matching filenames "
                       "will be scanned.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ","
        }
    },
    "exclude-files-pattern": {
        "short_name": "X",
        "description": "Python regex deny pattern. Matching filenames will "
                       "not be scanned.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ","
        }
    },
    "allow-io-errors": {
        "description": "Allow scanning to continue if IO errors are "
                       "encountered. Files that cannot be read will "
                       "be skipped and a warning will be logged. This is "
                       "the default behavior.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True
    },
    "chunk-size": {
        "short_name": "z",
        "description": "Size of file chunks that will be scanned. Use a whole "
                       "number followed by one of the following suffixes: b "
                       "(byte), k (kibibyte), m (mebibyte). Defaults to 3m.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": 3145728,
        "meta": {
            "value_type": byte_length
        }
    },
    "scanned-content-limit": {
        "short_name": "M",
        "description": "The maximum amount of data to scan in each file."
                       " Content beyond this limit will not be scanned."
                       " Defaults to 50 mebibytes. Use a whole number followed"
                       " by one of the following suffixes: b (byte),"
                       " k (kibibyte), m (mebibyte).",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": byte_length('50m'),
        "meta": {
            "value_type": byte_length
        }
    },
    "match-engine": {
        "description": "The regex engine to use for malware scanning.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": MatchEngine.get_default_option(),
        "meta": {
            "valid_options": MatchEngine.get_options()
        }
    },
    "match-all": {
        "description": "If set, all possible signatures will be checked "
                       "against each scanned file. Otherwise, only the "
                       "first matching signature will be reported",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    },
    "pcre-backtrack-limit": {
        "description": "The regex backtracking limit for signature evaluation",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": PCRE_DEFAULT_MATCH_LIMIT,
        "meta": {
            "value_type": int
        }
    },
    "pcre-recursion-limit": {
        "description": "The regex recursion limit for signature evaluation",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": PCRE_DEFAULT_MATCH_LIMIT_RECURSION,
        "meta": {
            "value_type": int
        }
    },
    "workers": {
        "short_name": "w",
        "description": "Number of worker processes used to perform scanning. "
                       "Defaults to 1 worker process.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": 1
    },
    "progress": {
        "description": "Display scan progress in the terminal with a curses "
                       "interface",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False
    },
    "pre-compile": {
        "description": "Pre-compile and cache the signature set without "
                       "actually running a scan",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False,
        "category": "Signature Compilation"
    },
    "pre-compile-generic": {
        "description": "Pre-compile and cache the signature set without "
                       "any CPU-specific optimizations and without running "
                       "a scan",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False,
        "category": "Signature Compilation"
    },
    "pattern-database-path": {
        "description": "Use an alternate path for storage of the pattern "
                       "database",
        "context": "ALL",
        "argument_type": "OPTION",
        "meta": {
            "accepts_file": True
        },
        "default": None,
        "category": "Signature Compilation"
    },
    "compile-local": {
        "description": "Always compile the signature set locally rather than "
                       "attempting to download a pre-compiled set",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False,
        "category": "Signature Compilation"
    },
    "re-compile": {
        "description": "Always re-compile the signature set rather than using "
                       "a cached version (when compiling locally)",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False,
        "category": "Signature Compilation"
    },
    "profile": {
        "description": "Profile scan performance",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False,
        "hidden": True
    },
    "profile-path": {
        "description": "Path at which to save profiling results",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": None,
        "hidden": True
    },
    "direct-io": {
        "short_name": "D",
        "description": "Use direct IO when opening files to avoid caching",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    }
}


cacheable_types = {
    'wordfence.intel.signatures.SignatureSet',
    'wordfence.intel.signatures.CommonString',
    'wordfence.intel.signatures.Signature',
    'wordfence.intel.signatures.PrecompiledSignatureSet',
    'wordfence.api.licensing.License'
}

examples = [
    UsageExample(
        'Scan /var/www/html for malicious files and display real-time output',
        'wordfence malware-scan --progress /var/www/html'
    ),
    UsageExample(
        'Scan for malicious files in multiple directories and record the '
        'results to a CSV file',
        'wordfence malware-scan --output-path /home/user/wordfence-cli.csv '
        '/home/user-one/dir /home/user-two/dir /home/user-three/dir'
    ),
    UsageExample(
        'Scan a directory for malicious files with any extension using 4 '
        'worker processes, writing each matching file name to a line in '
        'results.txt',
        'wordfence malware-scan -a -w 4 --output-path /path/to/results.txt '
        '--output-format line-delimited /path/to/scan'
    ),
    UsageExample(
        'Scan only files with a .js extension for malware',
        'wordfence malware-scan --include-files-pattern ".*\\.js$" '
        '/path/to/scan'
    ),
    UsageExample(
        'Find files updated in the last 60 and scan them for malware',
        'find /var/www/ -cmin -60 -type f -print0 | wordfence malware-scan '
        '--output-path /home/username/wordfence-cli.csv'
    )
]

definition = SubcommandDefinition(
    name='malware-scan',
    usage='[OPTIONS] [PATH]...',
    description='Scan files for malware',
    config_definitions=config_definitions,
    config_section='MALWARE_SCAN',
    cacheable_types=cacheable_types,
    previous_names={'scan'},
    examples=examples,
    uses_license=True,
    accepts_directories=True,
    accepts_files=True
)



import shutil
import os
from typing import Dict, Optional, Any, List

from .config.config_items import ConfigItemDefinition, Context
from .subcommands import SubcommandDefinition


COMMAND = 'wordfence'


class OptionHelp:

    def __init__(
                self,
                long_name: str,
                short_name: Optional[str],
                description: str,
                category: str,
                default: str,
                valid_values: Optional[List[str]],
                context: Context,
                is_flag: bool = False,
            ):
        self.long_name = long_name
        self.short_name = short_name
        self.description = description
        self.category = category
        self.default = default
        self.valid_values = valid_values
        self.context = context
        self.is_flag = is_flag
        self.label = self.generate_label()

    def generate_label(self) -> str:
        if self.short_name is None:
            short = '   '
        else:
            short = f'-{self.short_name},'
        return f'{short} --{self.long_name}'


class LineFormatter:

    def __init__(self, terminal_size: os.terminal_size):
        self.terminal_size = terminal_size

    def split_line(
                self,
                line: str,
                max_length: int,
                offset: int = 0,
                first: bool = True
            ) -> List[str]:
        if offset > max_length:
            offset = 0
        lines = []
        while len(line) > 0:
            end = max_length - 1
            if len(line) > max_length:
                try:
                    next_break = line.rindex(' ', 0, end)
                except ValueError:
                    next_break = end
            else:
                next_break = len(line)
            next = line[:next_break]
            line = line[next_break + 1:]
            if not first:
                next = next.rjust(len(next) + offset)
            lines.append(next)
            if first:
                first = False
                max_length -= offset
        return lines

    def join_lines(
                self,
                lines: List[str],
                delimiter: str = '\n',
                offset: int = 0,
            ) -> str:
        final_lines = []
        max_length = self.terminal_size.columns
        for input_line in lines:
            initial = True
            for real_line in input_line.splitlines():
                if len(real_line) > max_length or not initial:
                    final_lines.extend(
                            self.split_line(
                                    real_line,
                                    max_length,
                                    offset,
                                    first=initial
                                )
                        )
                else:
                    final_lines.append(real_line)
                initial = False
        return delimiter.join(final_lines)

    def join_chunks(
                self,
                chunks: List[str],
                delimiter: str = '\n\n'
            ) -> str:
        return delimiter.join(chunks)


SPACER = '  '
SPACER_LENGTH = len(SPACER)


class OptionFormatter:

    def __init__(
                self,
                config_map: Dict[str, ConfigItemDefinition],
                terminal_size: os.terminal_size
            ):
        self.terminal_size = terminal_size
        self.categories = {}
        self.max_label_length = 0
        self._load_options(config_map)
        self.line_formatter = LineFormatter(terminal_size)

    def _add_option_help(self, option: OptionHelp) -> None:
        try:
            category = self.categories[option.category]
        except KeyError:
            category = {}
            self.categories[option.category] = category
        category[option.long_name] = option

    def _load_options(
                self,
                config_map: Dict[str, ConfigItemDefinition]
            ) -> None:
        for item in config_map.values():
            if item.hidden or item.context is Context.CONFIG:
                continue
            valid_values = None
            if item.has_options_list():
                valid_values = item.meta.valid_options
            option = OptionHelp(
                    item.name,
                    item.short_name,
                    item.description,
                    item.category,
                    item.default,
                    valid_values,
                    item.context,
                    item.is_flag()
                )
            self._add_option_help(option)
            self.max_label_length = max(
                    self.max_label_length,
                    len(option.label)
                )

    def _offset(self, string: str, offset: int) -> str:
        return string.rjust(offset + len(string))

    def format_category(
                self,
                title: str,
                options: Dict[str, OptionHelp]
            ) -> str:
        lines = [
                f'{title}:'
            ]
        offset = (SPACER_LENGTH * 2) + self.max_label_length
        for option in options.values():
            label = option.label.ljust(self.max_label_length)
            lines.append(
                    f'{SPACER}{label}{SPACER}{option.description}'
                )
            if option.valid_values is not None:
                valid_options = 'Options: '
                valid_options += ', '.join(option.valid_values)
                lines.append(self._offset(valid_options, offset))
            if option.is_flag and (
                        option.default or
                        option.default is None or
                        option.context is not Context.CLI
                    ):
                lines.append(self._offset(
                        f'(use --no-{option.long_name} to disable)',
                        offset
                    ))
            elif isinstance(option.default, str) and len(option.default):
                lines.append(self._offset(
                        f'(default: {option.default})',
                        offset
                    ))
        return self.line_formatter.join_lines(lines, offset=offset)

    def format_options(self) -> str:
        sections = []
        for title, options in self.categories.items():
            section = self.format_category(title, options)
            sections.append(section)
        return self.line_formatter.join_chunks(sections)


class HelpGenerator:

    def __init__(
                self,
                terminal_size: os.terminal_size
            ):
        self.terminal_size = terminal_size
        self.line_formatter = LineFormatter(terminal_size)

    def _generate_usage_details(self) -> List[str]:
        raise NotImplementedError()

    def generate_usage(self) -> str:
        details = self._generate_usage_details()
        prefix = '\n' if len(details) > 1 else ''
        usage_lines = []
        for line in details:
            usage_lines.append(f'{COMMAND} {line}')
        return prefix + '\n'.join(usage_lines)

    def _get_config_map(self) -> Dict[str, ConfigItemDefinition]:
        raise NotImplementedError()

    def generate_options(self) -> str:
        config_map = self._get_config_map()
        formatter = OptionFormatter(config_map, self.terminal_size)
        return formatter.format_options()

    def generate_description(self) -> Optional[str]:
        return None

    def generate_examples(self) -> Optional[str]:
        return None

    def generate_subcommands(self) -> Optional[str]:
        return None

    def generate_help(self) -> str:
        usage = self.generate_usage()
        description = self.generate_description()
        examples = self.generate_examples()
        options = self.generate_options()
        subcommands = self.generate_subcommands()
        sections = [
                f'Usage: {usage}'
            ]
        if description is not None and len(description) > 0:
            sections.append(description)
        if examples is not None and len(examples) > 0:
            sections.append(f'Examples:\n{examples}')
        if len(options) > 0:
            sections.append(options)
        if subcommands is not None and len(subcommands) > 0:
            sections.append(f'Subcommands:\n{subcommands}')
        return self.line_formatter.join_chunks(sections)

    def display_help(self) -> str:
        help = self.generate_help()
        print(help)


class BaseHelpGenerator(HelpGenerator):

    def __init__(
                self,
                config_map: Dict[str, ConfigItemDefinition],
                subcommand_definitions: Dict[str, SubcommandDefinition],
                terminal_size: os.terminal_size
            ):
        self.config_map = config_map
        self.subcommand_definitions = subcommand_definitions
        super().__init__(terminal_size)

    def _generate_usage_details(self) -> List[str]:
        return ['<SUBCOMMAND> [OPTIONS]']

    def _get_config_map(self) -> Dict[str, ConfigItemDefinition]:
        return self.config_map

    def generate_subcommands(self) -> Optional[str]:
        subcommands = {}
        max_name_length = 0
        for definition in self.subcommand_definitions.values():
            subcommands[definition.name] = definition.description
            max_name_length = max(len(definition.name), max_name_length)
        lines = []
        for name, description in subcommands.items():
            padded_name = name.ljust(max_name_length)
            lines.append(
                    f'{SPACER}{padded_name}{SPACER}{description}'
                )
        offset = (SPACER_LENGTH * 2) + max_name_length
        return self.line_formatter.join_lines(lines, offset=offset)


class SubcommandHelpGenerator(HelpGenerator):

    def __init__(
                self,
                definition: SubcommandDefinition,
                base_config_map: Dict[str, ConfigItemDefinition],
                terminal_size: os.terminal_size
            ):
        self.definition = definition
        self.base_config_map = base_config_map
        super().__init__(terminal_size)

    def _generate_usage_row(self, usage: str) -> str:
        return f'{self.definition.name} {usage}'

    def _generate_usage_details(self) -> List[str]:
        lines = []
        if isinstance(self.definition.usage, str):
            lines.append(self._generate_usage_row(self.definition.usage))
        else:
            for usage in self.definition.usage:
                lines.append(usage)
        return lines

    def _get_config_map(self) -> Dict[str, ConfigItemDefinition]:
        return {**self.base_config_map, **self.definition.get_config_map()}

    def generate_description(self) -> str:
        lines = [
                self.line_formatter.join_lines([self.definition.description])
            ]
        if self.definition.long_description is not None:
            lines.append(self.line_formatter.join_lines(
                        [self.definition.long_description]
                    )
                )
        return self.line_formatter.join_chunks(lines)

    def generate_examples(self) -> List[str]:
        lines = []
        if self.definition.examples is None:
            return lines
        index = 1
        for example in self.definition.examples:
            lines.append(f'{SPACER}{index}. {example.description}')
            lines.append(f'{SPACER}{SPACER}{SPACER}{example.command}')
            index += 1
        return self.line_formatter.join_lines(
                lines,
                offset=SPACER_LENGTH * 3
            )


class Helper:

    def __init__(
                self,
                subcommand_definitions: Dict[str, SubcommandDefinition],
                base_config_map: Dict[str, ConfigItemDefinition],
                terminal_size: Optional[os.terminal_size] = None
            ):
        self.subcommand_definitions = subcommand_definitions
        self.base_config_map = base_config_map
        self.terminal_size = terminal_size \
            if terminal_size is not None \
            else shutil.get_terminal_size()
        self.generators = {}

    def _initialize_generator(
                self,
                subcommand: Optional[str]
            ) -> HelpGenerator:
        if subcommand is None:
            return BaseHelpGenerator(
                    self.base_config_map,
                    self.subcommand_definitions,
                    self.terminal_size
                )
        else:
            try:
                definition = self.subcommand_definitions[subcommand]
                return SubcommandHelpGenerator(
                        definition,
                        self.base_config_map,
                        self.terminal_size
                    )
            except KeyError:
                raise ValueError(f'Invalid subcommand: {subcommand}')

    def get_generator(self, subcommand: Optional[str] = None) -> HelpGenerator:
        try:
            return self.generators[subcommand]
        except KeyError:
            generator = self._initialize_generator(subcommand)
            self.generators[subcommand] = generator
            return generator

    def _invoke_generator_method(
                self,
                subcommand: Optional[str],
                method_name: str
            ) -> Any:
        generator = self.get_generator(subcommand)
        method = getattr(generator, method_name)
        if method is None:
            raise ValueError(f'Invalid generator method: {method_name}')
        return method()

    def generate_usage(self, subcommand: Optional[str] = None) -> str:
        return self._invoke_generator_method(subcommand, 'generate_usage')

    def generate_help(self, subcommand: Optional[str] = None) -> str:
        return self._invoke_generator_method(subcommand, 'generate_help')

    def display_help(self, subcommand: Optional[str] = None) -> None:
        return self._invoke_generator_method(subcommand, 'display_help')



import requests
from typing import Optional
from urllib.parse import urlencode

from .licensing import License
from .exceptions import ApiException
from ..util.validation import Validator, ValidationException

DEFAULT_TIMEOUT = 30


class NocClient:

    def __init__(
                self,
                license: Optional[License] = None,
                base_url: str = None,
                timeout: int = DEFAULT_TIMEOUT
            ):
        self.license = license
        self.base_url = base_url \
            if base_url is not None \
            else self.get_default_base_url()
        self.timeout = timeout

    def get_default_base_url(self) -> str:
        raise ApiException('No default base URL is defined')

    def build_query(self, action: str, base_query: dict = None) -> dict:
        if base_query is None:
            query = {}
        else:
            query = base_query.copy()
        query['action'] = action
        if self.license is not None:
            query['k'] = self.license.key
        query['cli'] = 1
        return query

    def request(
                self,
                action: str,
                query: Optional[dict] = None,
                body: Optional[dict] = None,
                json: bool = True
            ):
        query = self.build_query(action, query)
        url = self.base_url + '?' + urlencode(query)
        try:
            if body is None:
                response = requests.get(url, timeout=self.timeout)
            else:
                response = requests.post(url, timeout=self.timeout, data=body)
            if json:
                return response.json()
            else:
                return response.content
        except Exception as error:
            raise ApiException('Request failed') from error

    def validate_response(self, response, validator: Validator) -> None:
        try:
            validator.validate(response)
        except ValidationException as exception:
            raise ApiException('Response validation failed') from exception



import sys
from typing import Optional, Union

from ..util.io import StreamReader


class IoManager:

    def __init__(
                self,
                read_stdin: Optional[bool],
                input_delimiter: Union[str, bytes],
                write_stdout: Optional[bool] = None,
                output_path: Optional[str] = None,
                encode_paths: bool = False,
                binary: bool = False
            ):
        self.read_stdin = read_stdin
        if binary:
            self.input_delimiter = input_delimiter \
                if isinstance(input_delimiter, bytes) \
                else input_delimiter.encode('utf-8')
        else:
            self.input_delimiter = input_delimiter \
                if isinstance(input_delimiter, str) \
                else input_delimiter.decode('utf-8')
        self.write_stdout = write_stdout
        self.output_path = output_path
        self.binary = binary

    def should_read_stdin(self) -> bool:
        if sys.stdin is None:
            return False
        if self.read_stdin is None:
            return not sys.stdin.isatty()
        else:
            return self.read_stdin

    def get_input_reader(self) -> StreamReader:
        if not hasattr(self, 'input_reader'):
            self.input_reader = StreamReader(
                    sys.stdin,
                    self.input_delimiter,
                    binary=self.binary
                )
        return self.input_reader

    def should_write_stdout(self) -> bool:
        if sys.stdout is None or self.write_stdout is False:
            return False
        return self.write_stdout or self.output_path is None



import importlib
from collections import namedtuple
from types import ModuleType
from typing import Optional, Dict, Set, List, Union

from .config.typing import ConfigDefinitions
from .config.config_items import config_definitions_to_config_map, \
        ConfigItemDefinition
from .context import CliContext

VALID_SUBCOMMANDS = {
        'configure',
        'malware-scan',
        'vuln-scan',
        'remediate',
        'count-sites',
        'db-scan',
        'help',
        'version',
        'terms'
    }


def map_subcommand_to_module_name(subcommand: str) -> str:
    return subcommand.replace('-', '')


def import_subcommand_module(
            subcommand: str,
            submodule: Optional[str] = None
        ) -> ModuleType:
    name = map_subcommand_to_module_name(subcommand)
    if submodule is None:
        submodule = name
    target = f'.{name}.{submodule}'
    return importlib.import_module(
            target,
            package='wordfence.cli'
        )


class Subcommand:

    def __init__(self, context: CliContext):
        self.context = context
        # Aliases for shorter access to context properties
        self.config = context.config
        self.cache = context.cache
        self.helper = context.helper
        self.prepare()

    def prepare(self) -> None:
        pass

    def invoke(self) -> int:
        return 0

    def terminate(self) -> None:
        pass

    def generate_exception_message(
                self,
                exception: BaseException
            ) -> Optional[str]:
        return None


UsageExample = namedtuple('UsageExample', ['description', 'command'])


class SubcommandDefinition:

    def __init__(
                self,
                name: str,
                usage: Union[str, List[str]],
                description: str,
                config_definitions: ConfigDefinitions,
                config_section: str,
                cacheable_types: Set[str],
                requires_config: bool = True,
                previous_names: Set[str] = None,
                examples: List[UsageExample] = None,
                uses_license: bool = False,
                accepts_files: bool = False,
                accepts_directories: bool = False,
                long_description: Optional[str] = None
            ):
        self.name = name
        self.usage = usage
        self.description = description
        self.config_definitions = config_definitions
        self.config_section = config_section
        self.config_map = None
        self.cacheable_types = cacheable_types
        self.requires_config = requires_config
        self.previous_names = previous_names if previous_names is not None \
            else set()
        self.examples = examples
        self.uses_license = uses_license
        self.accepts_files = accepts_files
        self.accepts_directories = accepts_directories
        self.long_description = long_description

    def get_config_map(self) -> Dict[str, ConfigItemDefinition]:
        if self.config_map is None:
            self.config_map = config_definitions_to_config_map(
                    self.config_definitions
                )
        return self.config_map

    def accepts_option(self, name: str) -> bool:
        return name in self.config_definitions

    def accepts_paths(self) -> bool:
        return self.accepts_files or self.accepts_directories

    def initialize_subcommand(self, context: CliContext) -> Subcommand:
        module = import_subcommand_module(self.name)
        assert hasattr(module, 'factory')
        assert callable(module.factory)
        subcommand = module.factory(context)
        assert isinstance(subcommand, Subcommand)
        return subcommand


def load_subcommand_definition(subcommand: str) -> SubcommandDefinition:
    module = import_subcommand_module(subcommand, 'definition')
    assert hasattr(module, 'definition')
    assert isinstance(module.definition, SubcommandDefinition)
    return module.definition


def load_subcommand_definitions() -> Dict[str, SubcommandDefinition]:
    definitions = dict()
    for subcommand in VALID_SUBCOMMANDS:
        definitions[subcommand] = load_subcommand_definition(subcommand)
    return definitions



import csv
import sys
import os
from typing import IO, List, Any, Callable, Iterable, Type, Dict, Optional, \
        Union
from enum import Enum
from contextlib import nullcontext
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
from email.headerregistry import Address
from socket import gethostname

from wordfence.logging import log
from wordfence.util.io import resolve_path
from wordfence.util.html import Document, Tag, Stylesheet, Style, RawHtml, \
        HtmlContent
from .context import CliContext
from .io import IoManager
from .email import Mailer


class ReportingException(Exception):
    pass


class ReportColumn:

    def __init__(
                self,
                header: str,
                extractor: Callable[[Any], str]
            ):
        self.header = header
        self.extractor = extractor

    def extract_value(self, record: Any) -> str:
        return self.extractor(record)


class ReportColumnEnum(ReportColumn, Enum):

    def __init__(
                self,
                header: str,
                extractor: Callable[[Any], str]
            ):
        super().__init__(header, extractor)

    @classmethod
    def get_options(cls) -> List[str]:
        return [column.header for column in cls]

    @classmethod
    def get_options_as_string(cls, delimiter: str = ', ') -> str:
        return ', '.join(cls.get_options())

    @classmethod
    def for_option(cls, header: str) -> ReportColumn:
        for column in cls:
            if column.header == header:
                return column
        raise ValueError(f'Unrecognized report column: {header}')


REPORT_COLUMNS_ALL = 'all'


class ReportWriter:

    def __init__(self, target: IO):
        self._target = target
        self.initialize()

    def initialize(self) -> None:
        pass

    def write_row(self, data: List[str]):
        pass

    def allows_headers(self) -> bool:
        return True

    def allows_column_customization(self) -> bool:
        return True

    def get_max_columns(self) -> Optional[int]:
        return None


class CsvReportWriter(ReportWriter):

    def initialize(self):
        self.writer = csv.writer(self._target, delimiter=self.get_delimiter())

    def get_delimiter(self) -> str:
        return ','

    def write_row(self, data: List[str]) -> None:
        self.writer.writerow(data)


class TsvReportWriter(CsvReportWriter):

    def get_delimiter(self) -> str:
        return '\t'


class SingleColumnWriter(ReportWriter):

    def __init__(self, target: IO, delimiter: str):
        super().__init__(target)
        self.delimiter = delimiter

    def write_row(self, data: List[str]) -> None:
        for index, value in enumerate(data):
            if index > 0:
                break
            if value is None:
                value = ''
            self._target.write(str(value) + self.delimiter)

    def get_max_columns(self) -> Optional[int]:
        return 1


class RowlessWriter(ReportWriter):

    def allows_headers(self) -> bool:
        return False

    def allows_column_customization(self) -> bool:
        return False

    def write_row(self, data: List[str]) -> None:
        pass

    def write_record(self, record) -> None:
        raise NotImplementedError()


class BaseHumanReadableWriter(RowlessWriter):

    def format_record(self, record) -> str:
        raise NotImplementedError()

    def write_record(self, record) -> None:
        self._target.write(self.format_record(record))
        self._target.write('\n')


class ReportFormat:

    def __init__(
                self,
                option: str,
                initializer: Callable[[IO, List[ReportColumn]], ReportWriter],
                allows_headers: bool = True,
                allows_column_customization: bool = True
            ):
        self.option = option
        self.initializer = initializer
        self.allows_headers = allows_headers
        self.allows_column_customization = allows_column_customization

    def initialize_writer(
                self,
                stream: IO,
                columns: List[ReportColumn]
            ) -> ReportWriter:
        return self.initializer(stream, columns)


REPORT_FORMAT_CSV = ReportFormat(
        'csv',
        lambda stream, columns: CsvReportWriter(stream)
    )
REPORT_FORMAT_TSV = ReportFormat(
        'tsv',
        lambda stream, columns: TsvReportWriter(stream)
    )
REPORT_FORMAT_NULL_DELIMITED = ReportFormat(
        'null-delimited',
        lambda stream, columns: SingleColumnWriter(stream, "\0")
    )
REPORT_FORMAT_LINE_DELIMITED = ReportFormat(
        'line-delimited',
        lambda stream, columns: SingleColumnWriter(stream, "\n")
    )


class ReportFormatEnum(Enum):

    @classmethod
    def get_options(cls) -> List[str]:
        return [format.value.option for format in cls]

    @classmethod
    def for_option(cls, option: str):
        for format in cls:
            if format.value.option == option:
                return format
        raise ValueError(f'Unrecognized report format: {option}')


class ReportRecord:
    pass


def generate_html_table(results: Dict[str, Any]) -> Tag:
    table = Tag(
            'table',
            {
                'class': 'results',
                'align': 'center',
            }
        )
    for key, value in results.items():
        table.append(
            Tag('tr')
            .append(Tag('th', {'align': 'left'})
                    .append(key))
            .append(Tag('td', {'align': 'right'})
                    .append(str(value)))
        )
    return table


def generate_report_email_html(
            content: HtmlContent,
            title: str,
            hostname: str
        ) -> Document:
    document = Document()

    styles = Stylesheet()

    styles.add(
            Style(
                'th, td',
                {
                    'padding': '8px'
                }
            ),
            Style(
                'h1.logo',
                {
                    'font-family': 'serif',
                    'font-size': '18px'
                }
            ),
            Style(
                '.cli-green',
                {
                    'color': '#008000'
                }
            ),
            Style(
                '.logo .cli',
                {
                    'font-family': 'sans'
                }
            ),
            Style(
                'h2',
                {
                    'font-size': '16px'
                }
            ),
            Style(
                'table.container td',
                {
                    'text-align': 'center',
                }
            ),
            Style(
                'div.content',
                {
                    'min-width': '600px',
                    'width': '800px',
                    'background-color': '#F0F0F0',
                    'font-size': '12px',
                    'padding': '8px'
                }
            ),
            Style(
                'table.results, table.results td, table.results th',
                {
                    'border': '1px solid black',
                    'border-collapse': 'collapse'
                }
            ),
            Style(
                'table.results tr, table.results, td',
                {
                    'font-size': '12px'
                }
            )
        )

    document.head.append(styles)

    content_section = Tag('div', {'class': 'content'})
    container = Tag(
                    'table',
                    {
                        'class': 'container',
                        'align': 'center',
                        'cellspacing': '0',
                        'cellpadding': '0',
                        'border': '0'
                    }
                ).append(
                    Tag('tr')
                    .append(
                        Tag('td', {'align': 'center'}).append(content_section)
                    )
                )

    header = Tag('div')
    header.append(
            Tag('h1', {'class': 'logo'})
            .append(RawHtml(
                '<span class="cli-green">Word</span>fence '
                '<span class="cli">CLI</span>'))
        )
    header.append(
            Tag('h2')
            .append(title)
            .append(' for ')
            .append(Tag('font', {'face': 'monospace'})
                    .append(hostname))
        )
    content_section.append(header)

    content_section.append(content)

    document.body.append(container)
    return document


class ReportEmail:

    def __init__(
                self,
                recipient: Address,
                subject: str,
                plain_content: str,
                html_content: str
            ):
        self.recipient = recipient
        self.subject = subject
        self.plain_content = plain_content
        self.html_content = html_content

    def to_mime_multipart(self) -> MIMEMultipart:
        message = MIMEMultipart()
        message['Subject'] = self.subject

        content = MIMEMultipart('alternative')

        body_plain = MIMEText(self.plain_content, 'plain')
        content.attach(body_plain)

        body_html = MIMEText(self.html_content, 'html')
        content.attach(body_html)

        message.attach(content)

        return message


class Report:

    def __init__(
                self,
                format: ReportFormat,
                columns: List[ReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool = False
            ):
        self.format = format.value
        self.columns = columns
        self.email_addresses = email_addresses
        self.mailer = mailer
        self.write_headers = write_headers
        self.headers_written = False
        self.writers = []
        self.has_custom_columns = False
        self.files = {}
        self.rows_written = 0

    def add_target(self, stream: IO, filename: Optional[str] = None) -> None:
        if filename is not None:
            self.files[filename] = stream
        writer = self.format.initialize_writer(stream, self.columns)
        if self.write_headers and not writer.allows_headers():
            log.warning(
                    'Headers are not supported when using the '
                    f'{self.format.option} output format'
                )
        if self.has_custom_columns \
                and not writer.allows_column_customization():
            log.warning(
                    'Columns cannot be specified when using the '
                    f'{self.format.option} output format'
                )
        max_columns = writer.get_max_columns()
        if max_columns is not None and len(self.columns) > max_columns:
            log.warning(
                    'Too many output columns requested for specified format '
                    f'(only {max_columns} allowed)'
                )
        self.writers.append(writer)

    def _write_row(self, data: List[str], record: ReportRecord):
        self.rows_written += 1
        for writer in self.writers:
            if isinstance(writer, RowlessWriter):
                writer.write_record(record)
            else:
                writer.write_row(data)

    def _write_headers(self) -> None:
        if self.headers_written or not self.write_headers:
            return
        headers = [column.header for column in self.columns]
        for writer in self.writers:
            if writer.allows_headers():
                writer.write_row(headers)
        self.headers_written = True

    def _format_record(self, record: ReportRecord) -> List[str]:
        row = []
        for column in self.columns:
            row.append(column.extract_value(record))
        return row

    def _write_record(self, record: ReportRecord) -> None:
        self._write_row(self._format_record(record), record)

    def write_records(self, records: Iterable[ReportRecord]) -> None:
        self._write_headers()
        for record in records:
            self._write_record(record)

    def write_record(self, record: ReportRecord) -> None:
        self._write_headers()
        self._write_record(record)

    def has_writers(self) -> bool:
        return len(self.writers) > 0

    def generate_email(
                self,
                recipient: str,
                attachments: Dict[str, str],
                hostname: str
            ) -> ReportEmail:
        raise NotImplementedError(
                'This report does not support email generation'
            )

    def send_emails(self) -> None:
        attachments = {}
        for name, file in self.files.items():
            file.seek(0)
            content = file.read()
            attachments[os.fsdecode(name)] = content
        hostname = gethostname()
        for recipient in self.email_addresses:
            recipient = Address(addr_spec=recipient)
            email = self.generate_email(recipient, attachments, hostname)
            email = email.to_mime_multipart()
            email['To'] = str(recipient)

            for name, content in attachments.items():
                attachment = MIMEApplication(
                        content,
                        Name=name
                    )
                attachment.add_header(
                        'Content-Disposition',
                        'attachment',
                        filename=name
                    )
                email.attach(attachment)

            self.mailer.send(email)

    def complete(self) -> None:
        if self.rows_written > 0 and len(self.email_addresses) > 0:
            self.send_emails()


def get_config_options(
            formats: Type[ReportFormatEnum],
            columns: Type[ReportColumnEnum],
            default_columns: Union[List[ReportColumnEnum], str] =
            REPORT_COLUMNS_ALL,
            default_format: str = 'csv'
        ) -> Dict[str, Dict[str, Any]]:
    if not isinstance(default_columns, str):
        default_columns = ','.join(
                [column.header for column in default_columns]
            )
    header_formats = []
    column_formats = []
    for format in formats:
        if format.value.allows_headers:
            header_formats.append(format.value.option)
        if format.value.allows_column_customization:
            column_formats.append(format.value.option)
    header_format_string = ', '.join(header_formats)
    column_format_string = ', '.join(column_formats)
    return {
        "output": {
            "description": "Write results to stdout. This is the default "
                           "behavior when --output-path is not specified.",
            "context": "ALL",
            "argument_type": "OPTIONAL_FLAG",
            "default": None,
            "category": "Output Control"
        },
        "output-path": {
            "description": "Path to which to write results.",
            "context": "ALL",
            "argument_type": "OPTION",
            "default": None,
            "category": "Output Control",
            "meta": {
                "accepts_file": True
            }
        },
        "output-columns": {
            "description": ("An ordered, comma-delimited list of columns to"
                            " include in the output. Available columns: "
                            + columns.get_options_as_string()
                            + f"\nCompatible formats: {column_format_string}"
                            ),
            "context": "ALL",
            "argument_type": "OPTION",
            "default": default_columns,
            "meta": {
                "separator": ","
            },
            "category": "Output Control"
        },
        "output-format": {
            "short_name": "m",
            "description": "Output format used for result data.",
            "context": "ALL",
            "argument_type": "OPTION",
            "default": default_format,
            "meta": {
                "valid_options": formats.get_options()
            },
            "category": "Output Control"
        },
        "output-headers": {
            "description": "Include column headers in "
                           "output.\n"
                           f"Compatible formats: {header_format_string}",
            "context": "ALL",
            "argument_type": "FLAG",
            "default": None,
            "category": "Output Control"
        }
    }


class ReportManager:

    def __init__(
                self,
                formats: Type[ReportFormatEnum],
                columns: Type[ReportColumnEnum],
                context: CliContext,
                read_stdin: Optional[bool],
                input_delimiter: Union[str, bytes],
                binary_input: bool = False
            ):
        self.formats = formats
        self.columns = columns
        self.context = context
        self.config = context.config
        self.read_stdin = read_stdin
        self.input_delimiter = input_delimiter
        self.binary_input = binary_input
        self.email_addresses = [] if self.config.email is None \
            else self.config.email
        self.io_manager = None

    def will_email(self) -> bool:
        return len(self.email_addresses) > 0

    def get_config_options(self) -> Dict[str, Dict[str, Any]]:
        return get_config_options(
                self.formats,
                self.columns
            )

    def get_io_manager(self) -> IoManager:
        if self.io_manager is None:
            self.io_manager = IoManager(
                    self.read_stdin,
                    self.input_delimiter,
                    self.config.output,
                    self.config.output_path,
                    binary=self.binary_input
                )
        return self.io_manager

    def open_output_file(self) -> Optional[IO]:
        mode = 'w+' if self.will_email() else 'w'
        return open(resolve_path(self.config.output_path), mode) \
            if self.config.output_path is not None \
            else nullcontext()

    def _instantiate_report(
                self,
                format: ReportFormat,
                columns: List[ReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool
            ) -> Report:
        raise Exception(
                'Report instantiation must be implemented in a child class'
            )

    def _get_stdout_target(self) -> IO:
        return sys.stdout

    def _add_targets(
                self,
                report: Report,
                output_file: Optional[IO]
            ) -> None:
        if self.io_manager.should_write_stdout():
            report.add_target(self._get_stdout_target())
        if output_file is not None:
            report.add_target(
                    output_file,
                    os.path.basename(self.config.output_path)
                )

    def _get_configured_columns(self) -> List[ReportColumn]:
        columns = []
        for option in self.config.output_columns:
            if option == REPORT_COLUMNS_ALL:
                for column in self.columns:
                    columns.append(column)
            else:
                columns.append(self.columns.for_option(option))
        return columns

    def initialize_report(self, output_file: Optional[IO] = None) -> Report:
        format = self.formats.for_option(
                self.config.output_format
            )
        columns = self._get_configured_columns()
        mailer = self.context.get_mailer() if self.will_email() else None
        report = self._instantiate_report(
                format,
                columns,
                self.email_addresses,
                mailer,
                self.config.output_headers
            )
        report.has_custom_columns = self.config.is_specified('output_columns')
        self._add_targets(report, output_file)
        if not report.has_writers():
            raise ReportingException(
                    'Please specify an output file using the --output-path'
                    ' option or add --output to write results to standard '
                    'output'
                )
        return report



import sys

from wordfence.util.input import prompt_yes_no, InputException
from wordfence.util.caching import Cacheable, NoCachedValueException, \
        InvalidCachedValueException, DURATION_ONE_DAY
from wordfence.api.licensing import License, LicenseSpecific
from .context import CliContext
from .licensing import LicenseManager

TERMS_URL = \
    'https://www.wordfence.com/wordfence-cli-license-terms-and-conditions/'
TERMS_CACHE_KEY = 'terms'
ACCEPTANCE_CACHE_KEY = 'terms-accepted'
CACHEABLE_TYPES = {
        'wordfence.cli.terms_management.LicenseTermsAcceptance'
    }


class LicenseTermsAcceptance(LicenseSpecific):

    def __init__(self, license: License, accepted: bool = False):
        super().__init__(license)
        self.accepted = accepted


class TermsManager:

    def __init__(self, context: CliContext, license_manager: LicenseManager):
        self.context = context
        self.license_manager = license_manager

    def prompt_acceptance_if_needed(self, use_api: bool = True):
        try:
            acceptance = self.context.cache.get(ACCEPTANCE_CACHE_KEY)
            if acceptance is True:
                self.record_acceptance(remote=False)
                return
            if acceptance.accepted:
                return
        except (NoCachedValueException, InvalidCachedValueException):
            if use_api:
                client = self.context.get_noc1_client()
                client.ping_api_key()
                self.prompt_acceptance_if_needed(False)
                return
        self.prompt_acceptance(license=self.license_manager.check_license())

    def _cache_acceptance(
                self,
                license: License,
                accepted: bool = True
            ):
        self.context.cache.put(
                ACCEPTANCE_CACHE_KEY,
                LicenseTermsAcceptance(license, accepted)
            )

    def record_acceptance(
                self,
                license: License = None,
                accepted: bool = True,
                remote: bool = True
            ) -> None:
        if license is None:
            license = self.license_manager.check_license()
        if remote:
            client = self.context.create_noc1_client(license)
            client.record_toupp()
        self._cache_acceptance(license, accepted)

    def trigger_update(self, updated: bool, license: License):
        if updated:
            self.context.cache.remove(TERMS_CACHE_KEY)
        self.record_acceptance(
                license=license,
                accepted=not updated,
                remote=False
            )
        if updated:
            self.prompt_acceptance(license)

    def prompt_acceptance(self, license: License):
        if self.context.config.accept_terms:
            self.record_acceptance(license=license)
            return
        if license.paid:
            edition = ''
        else:
            edition = ' Free edition'
        terms_accepted = False
        try:
            terms_accepted = prompt_yes_no(
                f'Your access to and use of Wordfence CLI{edition} is '
                'subject to the updated Wordfence CLI License Terms and '
                f'Conditions set forth at {TERMS_URL}. By entering "y" and '
                'selecting Enter, you agree that you have read and accept the '
                'updated Wordfence CLI License Terms and Conditions.',
                default=False
            )
        except InputException:
            print(
                    'Wordfence CLI does not appear to be running interactively'
                    ' and cannot prompt for agreement to the license terms. '
                    'Please run Wordfence CLI in a terminal or use the '
                    '--accept-terms command line option instead.'
                )
        if terms_accepted:
            self.record_acceptance(license=license)
        else:
            print(
                    'You must accept the terms in order to continue using'
                    ' Wordfence CLI.'
                )
            sys.exit(1)

    def _fetch_terms(self) -> str:
        client = self.context.get_noc1_client()
        return client.get_terms()

    def get_terms(self) -> str:
        cacheable = Cacheable(
                TERMS_CACHE_KEY,
                self._fetch_terms,
                DURATION_ONE_DAY
            )
        return cacheable.get(self.context.cache)



import smtplib
from email.message import Message
from email.headerregistry import Address
from enum import Enum
from typing import Optional
from os import popen, getuid
from socket import gethostname
from pwd import getpwuid

from ..logging import log
from .config.config import Config


class EmailException(Exception):
    pass


class SmtpTlsMode(Enum):
    NONE = 'none'
    SMTPS = 'smtps'
    STARTTLS = 'starttls'


class Sender:

    def send(self, message: Message) -> None:
        pass

    def close(self) -> None:
        pass


class SmtpSender(Sender):

    def __init__(
                self,
                host: str,
                port: Optional[int] = None,
                tls_mode: SmtpTlsMode = SmtpTlsMode.STARTTLS,
                user: Optional[str] = None,
                password: Optional[str] = None
            ):
        smtp_type = smtplib.SMTP_SSL if tls_mode is SmtpTlsMode.SMTPS \
            else smtplib.SMTP
        port = 0 if port is None else port
        try:
            self.smtp = smtp_type(
                    host=host,
                    port=port
                )
            if tls_mode is SmtpTlsMode.STARTTLS:
                log.debug('Starting SMTP TLS...')
                self.smtp.starttls()
            if user is not None:
                log.debug(f'Authenticating with SMTP server as {user}...')
                self.smtp.login(user, password)
        except smtplib.SMTPException as e:
            raise EmailException('SMTP client creation failed') from e

    def send(self, message: Message) -> None:
        try:
            log.debug(f"Sending email via SMTP to {message['To']}...")
            self.smtp.send_message(message)
        except smtplib.SMTPException as e:
            raise EmailException('Sending email via SMTP failed') from e

    def close(self) -> None:
        self.smtp.quit()


class SendmailSender(Sender):

    def __init__(self, executable: str):
        self.executable = executable

    def send(self, message: Message):
        log.debug(f"Sending email via sendmail to {message['To']}...")
        command = f'{self.executable} -t -oi'
        try:
            sendmail = popen(command, 'w')
            sendmail.write(message.as_string())
            result = sendmail.close()
            if result is not None:
                raise EmailException(f'Sendmail exited with code: {result}')
        except Exception as e:
            raise EmailException('Sendmail invocation failed') from e


def initialize_sender(config: Config) -> Sender:
    if config.smtp_host is None:
        return SendmailSender(config.sendmail_path)
    else:
        return SmtpSender(
                config.smtp_host,
                config.smtp_port,
                SmtpTlsMode(config.smtp_tls_mode),
                config.smtp_user,
                config.smtp_password
            )


def generate_default_from_address(display_name: str) -> Address:
    username = getpwuid(getuid()).pw_name
    hostname = gethostname()
    address = Address(
            display_name=display_name,
            username=username,
            domain=hostname
        )
    return address


class Mailer(Sender):

    def __init__(
                self,
                config: Config,
            ):
        self.config = config
        self.sender = None
        self.from_address = None

    def get_sender(self) -> Sender:
        if self.sender is None:
            self.sender = initialize_sender(self.config)
        return self.sender

    def get_from_address(self) -> str:
        if self.from_address is None:
            address = self.config.email_from
            display_name = 'Wordfence CLI'
            if address is None:
                address = generate_default_from_address(display_name)
            else:
                address = Address(
                        display_name=display_name,
                        addr_spec=address
                    )
            self.from_address = str(address)
        return self.from_address

    def send(self, message: Message) -> None:
        message['From'] = self.get_from_address()
        self.get_sender().send(message)

    def close(self) -> None:
        if self.sender is not None:
            self.sender.close()



import requests
from packaging import version
from typing import Optional

from . import caching
from .caching import NoCachedValueException
from ..version import __version__
from ..logging import log

API = 'https://api.github.com/repos/wordfence/wordfence-cli/releases/latest'


class Version:

    @staticmethod
    def get_latest() -> Optional[str]:
        try:
            response = requests.get(API).json()
            if 'tag_name' in response.keys():
                latest = response['tag_name']
                if latest[0] == 'v':
                    latest = latest[1:]
                return latest
            else:
                return None
        except requests.exceptions.RequestException:
            return None

    @staticmethod
    def check(cache: caching.Cache):
        try:
            latest_version = caching.Cache.get(
                cache,
                'latest_version',
                86400  # Latest version is valid for 24 hours
            )
        except NoCachedValueException:
            latest_version = Version.get_latest()
            if latest_version is None:
                log.error('Unable to fetch the latest version. '
                          'The version you are using may be out of date!')
                return
            else:
                caching.Cache.put(cache, 'latest_version', latest_version)

        if latest_version is None:
            return

        if version.parse(__version__) < version.parse(latest_version):
            log.warning('A newer version of the Wordfence CLI is available! '
                        'Updating to ' + latest_version + ' is recommended.')



import curses
import logging
import signal
import os
from typing import List, Optional, Deque
from logging import Handler
from collections import deque, namedtuple
from time import sleep

from wordfence.scanning.scanner import (ScanProgressUpdate, ScanMetrics,
                                        default_scan_finished_handler)
from ..banner.banner import get_welcome_banner
from ...util import timing
from ...util.unicode import filter_control_characters
from ...util.units import scale_byte_unit


class ProgressException(Exception):
    pass


_displays = []

METRIC_BOX_WIDTH = 39
"""
Hard-coded width of metric boxes

The actual width taken up will be the hard-coded value +2 to account for the
left and right borders. Each box on the same row will be separated by the
padding value as well.
"""


def reset_terminal() -> None:
    for display in _displays:
        display.end()


def resize_terminal(signalnum, frame) -> None:
    for display in _displays:
        display.queue_resize()


signal.signal(signal.SIGWINCH, resize_terminal)


Position = namedtuple('Position', ['y', 'x'])


class LayoutProperties:

    def __init__(
                self,
                lines: int,
                current_line: int,
                max_row_width: int
            ):
        self.lines = lines
        self.current_line = current_line
        self.max_row_width = max_row_width


class Box:

    def __init__(
                self,
                parent: Optional[curses.window] = None,
                border: bool = True,
                title: Optional[str] = None
            ):
        self.parent = parent
        self.border = border
        self.title = title
        self.window = None
        self.position = None
        self.last_size = None

    def _initialize_window(self, y: int = 0, x: int = 0) -> None:
        height, width = self.compute_size()
        if self.parent is None:
            self.window = curses.newwin(height, width, y, x)
        else:
            self.window = self.parent.subwin(height, width, y, x)
        self.position = Position(y, x)

    def set_position(self, y: int, x: int) -> None:
        if self.window is None:
            self._initialize_window(y, x)
        else:
            self.resize(1, 1)
            try:
                self.window.erase()
                self.window.mvderwin(y, x)
                self.window.mvwin(y, x)
                self.position = Position(y, x)
            except Exception as e:
                size = os.get_terminal_size()
                raise ValueError(
                        f"error moving window: y: {y}, x: {x}; "
                        f"height: {self.get_height()}; "
                        f"width: {self.get_width()}; "
                        f"lines: {size.lines}; "
                        f"columns: {size.columns}"
                    ) from e
            self.resize()

    def _require_window(self) -> None:
        if self.window is None:
            self._initialize_window()

    def compute_size(self) -> (int, int):
        height = self.get_height()
        width = self.get_width()
        if self.border:
            width += 2
            height += 2
        self.last_size = (height, width)
        return self.last_size

    def resize(
                self,
                lines: Optional[int] = None,
                cols: Optional[int] = None
            ) -> None:
        if self.window is None:
            return
        height, width = self.compute_size()
        if lines is not None:
            height = lines
        if cols is not None:
            width = cols
        self.window.erase()
        try:
            self.window.resize(height, width)
        except Exception:
            pass  # Ignore temporary errors during resizing
        self.update()

    def set_title(self, title: str) -> None:
        self.title = title

    def render(self) -> None:
        self._require_window()
        height, width = self.compute_size()
        if self.border:
            self.window.border()
        if self.title is not None:
            title_length = len(self.title)
            title_offset = 0
            if title_length < width:
                title_offset = int((width - title_length) / 2)
            try:
                self.window.addstr(0, title_offset, self.title)
            except Exception:
                pass  # Ignore temporary errors during resizing
        try:
            self.draw_content()
        except Exception:
            pass  # Ignore temporary errors during resizing

    def get_border_offset(self) -> int:
        return 1 if self.border else 0

    def draw_content(self) -> None:
        pass

    def update(self) -> None:
        self.render()
        self.window.syncup()
        self.window.noutrefresh()

    def resize_for_layout(self, properties: LayoutProperties) -> False:
        return False


class Metric:

    def __init__(self, label: str, value):
        self.label = label
        self.value = str(value)


class MetricBox(Box):

    def __init__(
                self,
                metrics: List[Metric],
                title: Optional[str] = None,
                parent: Optional[curses.window] = None
            ):
        self.metrics = metrics
        super().__init__(parent, title=title)

    def get_width(self) -> int:
        return METRIC_BOX_WIDTH

    def get_height(self) -> int:
        return len(self.metrics)

    def draw_content(self) -> None:
        offset = self.get_border_offset()
        width = self.get_width()
        for index, metric in enumerate(self.metrics):
            line = index + offset
            label = f'{metric.label}:'
            value_string = metric.value.rjust(width - len(label))
            self.window.addstr(line, offset, label + value_string)


class BannerBox(Box):

    def __init__(
                self,
                banner,
                parent: Optional[curses.window] = None
            ):
        self.banner = banner
        super().__init__(parent, border=False)

    def get_width(self):
        return self.banner.column_count

    def get_height(self):
        return self.banner.row_count

    def draw_content(self):
        offset = self.get_border_offset()
        for index, row in enumerate(self.banner.rows):
            self.window.addstr(index + offset, offset, row)


DEFAULT_MAX_MESSAGES = 512


class LogBox(Box):

    def __init__(
                self,
                columns: int,
                lines: int,
                max_messages: int = 0,
                parent: Optional[curses.window] = None
            ):
        self.columns = columns
        self.lines = lines
        self.messages = deque(
                maxlen=self._determine_max_messages(max_messages)
            )
        self.cursor_position = None
        super().__init__(parent, border=True)

    def _determine_max_messages(self, max_messages: int = 0) -> Optional[int]:
        if max_messages < 0:
            return None
        elif max_messages == 0:
            return max(self.lines, DEFAULT_MAX_MESSAGES)
        else:
            return max_messages

    def get_width(self):
        return self.columns

    def get_height(self):
        return self.lines

    def _map_messages_to_lines(self, offset: int) -> Deque[str]:
        lines = deque(maxlen=self.lines)
        remaining_lines = self.lines
        for message in reversed(self.messages):
            if remaining_lines == 0:
                break
            message_lines = []
            while len(message):
                if remaining_lines == 0:
                    break
                line = message[:self.columns]
                message = message[self.columns:]
                message_lines.append(line)
                remaining_lines -= 1
            for line in reversed(message_lines):
                lines.appendleft(line)
        return lines

    def draw_content(self) -> None:
        offset = self.get_border_offset()
        line_number = offset
        last_line_number = line_number
        last_line_length = 0
        for line in self._map_messages_to_lines(offset):
            last_line_number = line_number
            last_line_length = len(line)
            line = line.ljust(self.columns)
            try:
                self.window.addstr(line_number, offset, line)
            except Exception:
                break
            line_number += 1
        self.cursor_offset = Position(last_line_number, last_line_length)

    def add_message(self, message: str) -> None:
        self.messages.append(filter_control_characters(message))
        self.update()

    def get_cursor_position(self) -> Position:
        y = 0
        x = 0
        if self.position is not None:
            y += self.position.y
            x += self.position.x
        if self.cursor_offset is not None:
            y += self.cursor_offset.y
            x += self.cursor_offset.x
        return Position(y, x)

    def resize_for_layout(self, properties: LayoutProperties) -> bool:
        self.columns = properties.max_row_width - 2
        self.lines = properties.lines - properties.current_line - 2
        self.cursor_position = None
        if self.lines < 3:
            raise ProgressException(
                    'Insufficient space available to display log messages'
                )
        return True


class LogBoxHandler(Handler):

    def __init__(self, log_box: LogBox):
        self.log_box = log_box
        Handler.__init__(self)

    def emit(self, record):
        self.log_box.add_message(record.getMessage())
        pass


class LogBoxStream():

    def __init__(self, log_box: LogBox):
        self.log_box = log_box

    def write(self, line):
        self.log_box.add_message(line)


class BoxLayout:

    def __init__(self, lines: int, cols: int, padding: int = 1):
        self.lines = lines
        self.cols = cols
        self.padding = padding
        self.current_line = 0
        self._content = []
        self._unpositioned = []
        self.max_row_width = 0

    def add_box(self, box: Box) -> None:
        self._content.append(box)
        self._unpositioned.append(box)

    def add_break(self) -> None:
        self._content.append(None)
        self._unpositioned.append(None)

    def get_layout_properties(self) -> LayoutProperties:
        return LayoutProperties(
                    lines=self.lines,
                    current_line=self.current_line,
                    max_row_width=self.max_row_width
                )

    def _position_row(self, row: list) -> list:
        positioned = []
        extra = []
        row_width = 0
        unpadded_row_width = 0
        row_height = 0
        for box in row:
            box.resize_for_layout(self.get_layout_properties())
            height, width = box.compute_size()
            required_width = width + self.padding
            if len(positioned) and (
                        len(extra) or
                        row_width + required_width > self.cols
                    ):
                extra.append(box)
            else:
                row_width += required_width
                if row_width > self.cols:
                    raise ProgressException('Insufficient columns available')
                unpadded_row_width += width
                row_height = max(row_height, height)
                positioned.append((box, height, width))
        if self.current_line + row_height > self.lines:
            raise ProgressException('Insufficient lines available')
        box_count = len(positioned)
        padding = int((self.cols - unpadded_row_width) / (box_count + 1))
        padded_width = unpadded_row_width + padding * (box_count + 1)
        x = padding + int((self.cols - padded_width) / 2)
        final_row_width = 0
        previous_padding = 0
        for (box, height, width) in positioned:
            final_row_width += previous_padding
            y = self.current_line + round((row_height - height) / 2)
            box.set_position(y, x)
            x += width + padding
            final_row_width += width
            previous_padding = padding
        self.current_line += row_height + self.padding
        self.max_row_width = max(self.max_row_width, final_row_width)
        return extra

    def position(self) -> None:
        row = []
        items = self._unpositioned
        for item in items:
            if item is None:
                row = self._position_row(row)
            else:
                row.append(item)
        while len(row):
            row = self._position_row(row)
        self._unpositioned = []

    def update_content(self) -> None:
        for item in self._content:
            if item is not None:
                item.update()

    def reset(self) -> None:
        self.current_line = 0
        self.max_row_width = 0
        self._unpositioned = self._content.copy()

    def resize(self, lines: int, cols: int) -> None:
        self.lines = lines
        self.cols = cols
        self.reset()
        self.position()


class ProgressDisplay:

    METRICS_PADDING = 1
    METRICS_COUNT = 5
    MIN_MESSAGE_BOX_HEIGHT = 4

    def __init__(self, worker_count: int):
        _displays.append(self)
        self.worker_count = worker_count
        self.results_message = None
        self.pending_resize = False
        self._setup_curses()

    def _setup_curses(self) -> None:
        self.stdscr = curses.initscr()
        curses.noecho()
        curses.curs_set(0)
        self.terminal_size = os.get_terminal_size()
        self._initialize_content(self.terminal_size)

    def _initialize_content(self, size: os.terminal_size) -> None:
        self.clear()
        self.banner_box = self._initialize_banner()
        self.metric_boxes = self._initialize_metric_boxes()
        self.log_box = self._initialize_log_box()
        self.layout = self._initialize_layout(size)
        self.refresh()

    def clear(self):
        self.stdscr.clear()

    def refresh(self):
        self.stdscr.noutrefresh()
        curses.doupdate()

    def end_on_input(self):
        curses.flushinp()
        self.stdscr.nodelay(True)
        while True:
            key = self.stdscr.getch()
            if key != -1 and key != curses.KEY_RESIZE:
                break
            if self._resize_if_necessary():
                self._move_cursor_to_log_end()
            sleep(0.1)
        self.end()

    def end(self):
        curses.endwin()
        _displays.remove(self)

    def _initialize_banner(self) -> Optional[BannerBox]:
        banner = get_welcome_banner()
        if banner is None:
            return None
        return BannerBox(banner=banner, parent=self.stdscr)

    def _compute_rate(self, value: int, elapsed_time: float) -> int:
        if elapsed_time > 0:
            return int(value / elapsed_time)
        return 0

    def _get_metrics(
                self,
                update: ScanProgressUpdate,
                worker_index: Optional[int] = None
            ) -> List[Metric]:
        file_count = update.metrics.get_int_metric('counts', worker_index)
        byte_count = update.metrics.get_int_metric('bytes', worker_index)
        byte_value = scale_byte_unit(byte_count)
        match_count = update.metrics.get_int_metric('matches', worker_index)
        file_rate = self._compute_rate(file_count, update.elapsed_time)
        byte_rate = self._compute_rate(byte_count, update.elapsed_time)
        byte_rate = scale_byte_unit(byte_rate)
        metrics = [
                Metric('Files Processed', file_count),
                Metric('Bytes Processed', byte_value),
                Metric('Matches Found', match_count),
                Metric('Files / Second', file_rate),
                Metric('Bytes / Second', byte_rate)
            ]
        if len(metrics) > self.METRICS_COUNT:
            raise ValueError("Metrics count is out of sync")
        return metrics

    def _initialize_metric_boxes(self) -> List[MetricBox]:
        default_metrics = ScanMetrics(self.worker_count)
        default_update = ScanProgressUpdate(
                elapsed_time=0,
                metrics=default_metrics
            )
        boxes = []
        for index in range(0, self.worker_count + 1):
            if index == 0:
                worker_index = None
                title = 'Summary'
            else:
                worker_index = index - 1
                title = f'Worker {index}'
            box = MetricBox(
                    self._get_metrics(default_update, worker_index),
                    title=title,
                    parent=self.stdscr
                )
            boxes.append(box)
        return boxes

    def _initialize_log_box(self) -> LogBox:
        log_box = LogBox(
                    # Lines and columns are dynamic
                    columns=10,
                    lines=5,
                    parent=self.stdscr
                )
        return log_box

    def _initialize_layout(self, size: os.terminal_size) -> BoxLayout:
        layout = BoxLayout(size.lines, size.columns, self.METRICS_PADDING)
        if self.banner_box is not None:
            layout.add_box(self.banner_box)
        for index, box in enumerate(self.metric_boxes):
            layout.add_box(box)
            if index == 0:
                layout.add_break()
        layout.add_break()
        layout.add_box(self.log_box)
        layout.position()
        layout.update_content()
        return layout

    def _display_metrics(self, update: ScanProgressUpdate) -> None:
        for index in range(0, self.worker_count + 1):
            box = self.metric_boxes[index]
            worker_index = None if index == 0 else index - 1
            box.metrics = self._get_metrics(update, worker_index)
            box.update()

    def handle_update(self, update: ScanProgressUpdate) -> None:
        self._resize_if_necessary()
        try:
            self._display_metrics(update)
            self.refresh()
        except Exception as e:
            reset_terminal()
            raise ProgressException('Rendering progress update failed') from e

    def queue_resize(self) -> None:
        self.pending_resize = True

    def resize(self) -> None:
        size = os.get_terminal_size()
        smaller = size.columns < self.terminal_size.columns
        self.terminal_size = size
        if smaller:
            self.layout.resize(size.lines, size.columns)
        curses.resizeterm(size.lines, size.columns)
        self.stdscr.erase()
        self.stdscr.refresh()
        self.stdscr.resize(size.lines, size.columns)
        if not smaller:
            self.layout.resize(size.lines, size.columns)
        self.layout.update_content()
        self.refresh()

    def _resize_if_necessary(self) -> bool:
        if not self.pending_resize:
            return False
        try:
            self.resize()
            self.pending_resize = False
            return True
        except Exception as e:
            reset_terminal()
            raise ProgressException(
                    'Failed to adjust progress output to new terminal size'
                ) from e

    def get_log_handler(self) -> logging.Handler:
        return LogBoxHandler(self.log_box)

    def get_output_stream(self) -> LogBoxStream:
        return LogBoxStream(self.log_box)

    def _move_cursor_to_log_end(self) -> None:
        cursor_position = self.log_box.get_cursor_position()
        if cursor_position is not None:
            try:
                self.stdscr.move(cursor_position.y, cursor_position.x + 1)
            except Exception:
                pass

    def scan_finished_handler(
                self, metrics: ScanMetrics,
                timer: timing.Timer
            ) -> None:
        messages = default_scan_finished_handler(metrics, timer)
        self.results_message = messages.results
        self.log_box.add_message('Scan completed! Press any key to exit.')
        self._move_cursor_to_log_end()
        curses.curs_set(1)



import time


def unit_nanoseconds(ns: int) -> int:
    return ns


def unit_seconds(ns: int) -> int:
    return ns / 1000000000


def unit_milliseconds(ns: int) -> int:
    return ns / 1000000


class Timer:

    def __init__(self, start: bool = True):
        if start:
            self.start()
        else:
            self.start_time = None
        self.end_time = None
        self.previous_time = 0

    def _capture_time(self) -> int:
        return time.monotonic_ns()

    def start(self) -> None:
        self.start_time = self._capture_time()
        self.end_time = None

    def reset(self) -> None:
        self.start()

    def stop(self) -> None:
        self.end_time = self._capture_time()

    def resume(self) -> None:
        if self.start_time is not None:
            self.previous_time += self.get_elapsed(
                    unit=unit_nanoseconds,
                    total=False
                )
        self.start()

    def _get_current_elapsed(self) -> int:
        if self.start_time is None:
            return 0
        end_time = \
            self.end_time if self.end_time is not None \
            else self._capture_time()
        return end_time - self.start_time

    def get_elapsed(self, unit=unit_seconds, total: bool = True) -> int:
        previous_time = self.previous_time if total else 0
        return unit(previous_time + self._get_current_elapsed())



import json
import re
import base64
from typing import Callable, Optional

from .noc_client import NocClient
from .exceptions import ApiException
from .licensing import License

from ..intel.signatures import CommonString, Signature, SignatureSet, \
    PrecompiledSignatureSet, deserialize_precompiled_signature_set
from ..intel.database_rules import DatabaseRuleSet, JSON_VALIDATOR as \
    DATABASE_RULES_JSON_VALIDATOR, parse_database_rules
from ..util.validation import DictionaryValidator, ListValidator, Validator, \
    OptionalValueValidator
from ..util.platform import Platform

NOC1_BASE_URL = 'https://noc1.wordfence.com/v2.27/'


API_ERROR_PATTERN = re.compile(r'\{.*errorMsg')


class Client(NocClient):

    def get_default_base_url(self) -> str:
        return NOC1_BASE_URL

    def _generate_site_stats(self) -> str:
        return json.dumps({})

    def build_query(self, action: str, base_query: dict = None) -> dict:
        query = super().build_query(action, base_query)
        query['s'] = self._generate_site_stats()
        return query

    def register_terms_update_hook(
                self,
                callable: Callable[[bool, License], None]
            ) -> None:
        if not hasattr(self, 'terms_update_hooks'):
            self.terms_update_hooks = []
        self.terms_update_hooks.append(callable)

    def _trigger_terms_update_hooks(
                self,
                updated: bool,
                license: License
            ) -> None:
        if not hasattr(self, 'terms_update_hooks'):
            return
        for hook in self.terms_update_hooks:
            hook(updated, license)

    def register_license_update_hook(
                self,
                callable: Callable[[License], None]
            ) -> None:
        if not hasattr(self, 'license_update_hooks'):
            self.license_update_hooks = []
        self.license_update_hooks.append(callable)

    def _trigger_license_update_hooks(self, license: License) -> None:
        if not hasattr(self, 'license_update_hooks'):
            return
        for hook in self.license_update_hooks:
            hook(license)

    def _check_error_message(self, response: dict) -> None:
        if 'errorMsg' in response:
            raise ApiException(
                    'Error message received in response body',
                    response['errorMsg']
                )

    def validate_response(self, response, validator: Validator) -> None:
        if isinstance(response, dict):
            self._check_error_message(response)
            paid = bool('_isPaidKey' in response and response['_isPaidKey'])
            if paid != self.license.paid:
                self.license.paid = paid
                self._trigger_license_update_hooks(self.license)
            terms_updated = '_termsUpdated' in response
            self._trigger_terms_update_hooks(terms_updated, self.license)
        return super().validate_response(response, validator)

    def process_simple_request(self, action: str) -> bool:
        response = self.request(action)
        validator = DictionaryValidator({
                'ok': int
            })
        self.validate_response(response, validator)
        return bool(response['ok'])

    def get_patterns(self) -> dict:
        patterns = self.request('get_patterns')
        validator = DictionaryValidator({
            'badstrings': ListValidator(str),
            'commonStrings': ListValidator(str),
            'rules': ListValidator(ListValidator({
                0: int,
                1: int,
                2: str,
                3: str,
                4: str,
                5: int,
                6: str,
                7: str,
                8: ListValidator(int)
            })),
            'signatureUpdateTime': int,
            'word1': str,
            'word2': str,
            'word3': str
        })
        self.validate_response(patterns, validator)
        return patterns

    def get_malware_signatures(self) -> SignatureSet:
        patterns = self.get_patterns()
        common_strings = []
        signatures = {}
        for string in patterns['commonStrings']:
            common_strings.append(CommonString(string))
        for record in patterns['rules']:
            if record[5] != 0:
                continue
            signature_id = record[0]
            signatures[signature_id] = Signature(
                signature_id,
                record[2],
                record[7],
                record[3],
                record[8]
            )
            for index in record[8]:
                try:
                    common_strings[index].signature_ids.append(signature_id)
                except IndexError as index_error:
                    raise ApiException(
                            'Response data contains malformed common string '
                            'association'
                        ) from index_error
        return SignatureSet(common_strings, signatures, self.license)

    def get_precompiled_patterns(
                self,
                platform: str,
                library_version: str,
                library_type: Optional[str] = None,
                database_version: int = PrecompiledSignatureSet.VERSION
            ) -> dict:
        parameters = {
                'platform': platform,
                'library_version': library_version,
                'database_version': database_version
            }
        if library_type is not None:
            parameters['library_type'] = library_type
        response = self.request('get_precompiled_patterns', parameters)
        validator = DictionaryValidator({
                'data': OptionalValueValidator(str)
            })
        self.validate_response(response, validator)
        return response

    def get_precompiled_malware_signatures(
                self,
                platform: Platform,
                library_version: str,
                library_type: Optional[str] = None,
                database_version: int = PrecompiledSignatureSet.VERSION
            ) -> Optional[PrecompiledSignatureSet]:
        response = self.get_precompiled_patterns(
                platform.key,
                library_version,
                library_type,
                database_version
            )
        data = response['data']
        if data is None:
            return None
        data = base64.b64decode(data)
        signature_set = deserialize_precompiled_signature_set(data)
        signature_set.assign_license(self.license)
        if isinstance(signature_set, PrecompiledSignatureSet):
            return signature_set
        raise ApiException(
                'Malformed signature set data received from Wordfence API'
            )

    def ping_api_key(self) -> bool:
        return self.process_simple_request('ping_api_key')

    def get_cli_api_key(self, accept_terms: bool = False) -> str:
        response = self.request(
                'get_cli_api_key',
                {'accept_terms': int(accept_terms)}
            )
        validator = DictionaryValidator({
                'apiKey': str
            })
        self.validate_response(response, validator)
        return response['apiKey']

    def record_toupp(self) -> bool:
        success = self.process_simple_request('record_toupp')
        if success:
            self.terms_updated = False
        return success

    def get_terms(self) -> str:
        response = self.request('get_terms')
        validator = DictionaryValidator({
                'terms': str
            })
        self.validate_response(response, validator)
        return response['terms']

    def request_raw(
                self,
                action: str,
                query: Optional[dict] = None,
                body: Optional[dict] = None
            ) -> bytes:
        response = self.request(
                action,
                query,
                body,
                json=False
            )
        try:
            response_string = response.decode('utf-8')
            if API_ERROR_PATTERN.match(response_string):
                json_data = json.loads(response)
                self._check_error_message(json_data)
        except (UnicodeError, json.JSONDecodeError):
            pass  # If the response isn't valid JSON, then there's no error
        return response

    def get_wp_file_content(
                self,
                type: str,
                path: str,
                coreVersion: str,
                name: Optional[str] = None,
                version: Optional[str] = None
            ) -> str:
        parameters = {
                'cType': type,
                'file': path,
                'v': coreVersion,
            }
        if name is not None:
            parameters['cName'] = name
        if version is not None:
            parameters['cVersion'] = version
        response = self.request_raw(
                'get_wp_file_content',
                body=parameters
            )
        return response

    def get_database_rules(self) -> DatabaseRuleSet:
        response = self.request('get_database_rules')
        validator = DictionaryValidator({
                'rules': DATABASE_RULES_JSON_VALIDATOR
            })
        self.validate_response(response, validator)
        return parse_database_rules(response['rules'], pre_validated=True)



import os

from typing import List, Set, Optional
from uuid import UUID

from ...intel.vulnerabilities import VulnerabilityIndex, \
        VulnerabilityScanner, VulnerabilityFilter, AlreadyScannedException, \
        is_cve_id
from ...api.intelligence import VulnerabilityFeedVariant
from ...util.caching import Cacheable, DURATION_ONE_DAY
from ...util.versioning import version_to_str
from ...wordpress.site import WordpressSite, WordpressStructureOptions, \
        WordpressLocator, WordpressException
from ...wordpress.plugin import PluginLoader, Plugin
from ...wordpress.theme import ThemeLoader, Theme
from ...logging import log
from ..subcommands import Subcommand
from .reporting import VulnScanReportManager
from .exceptions import VulnScanningConfigurationException


class VulnScanSubcommand(Subcommand):

    def _load_vulnerability_index(
                self,
                variant: VulnerabilityFeedVariant
            ) -> VulnerabilityIndex:
        def initialize_vulnerability_index() -> VulnerabilityIndex:
            client = self.context.get_wfi_client()
            vulnerabilities = client.fetch_vulnerability_feed(variant)
            return VulnerabilityIndex(vulnerabilities)
        vulnerability_index = Cacheable(
                f'vulnerability_index_{variant.path}',
                initialize_vulnerability_index,
                DURATION_ONE_DAY
            )
        return vulnerability_index.get(self.cache)

    def _scan_plugins(
                self,
                plugins: List[Plugin],
                scanner: VulnerabilityScanner,
                path: bytes,
                child_scan: bool = False
            ) -> None:
        if not child_scan:
            scanner.add_scan_path(path)
        for plugin in plugins:
            log.debug(
                    f'Plugin {plugin.slug}, version: ' +
                    version_to_str(plugin.version)
                )
            scanner.scan_plugin(plugin, path)

    def _scan_plugin_directory(
                self,
                directory: bytes,
                scanner: VulnerabilityScanner,
            ) -> None:
        loader = PluginLoader(
                directory,
                allow_io_errors=self.config.allow_io_errors
            )
        plugins = loader.load_all()
        self._scan_plugins(plugins, scanner, directory)

    def _scan_themes(
                self,
                themes: List[Theme],
                scanner: VulnerabilityScanner,
                path: bytes,
                child_scan: bool = False
            ) -> None:
        if not child_scan:
            scanner.add_scan_path(path)
        for theme in themes:
            log.debug(
                    f'Theme {theme.slug}, version: ' +
                    version_to_str(theme.version)
                )
            scanner.scan_theme(theme, path)

    def _scan_theme_directory(
                self,
                directory: bytes,
                scanner: VulnerabilityScanner
            ) -> None:
        loader = ThemeLoader(
                directory,
                allow_io_errors=self.config.allow_io_errors
            )
        themes = loader.load_all()
        self._scan_themes(themes, scanner, directory)

    def _scan(
                self,
                path: bytes,
                scanner: VulnerabilityScanner,
                check_extensions: bool = False,
                structure_options: WordpressStructureOptions = None,
                scan_path: Optional[bytes] = None
            ) -> None:
        scanner.add_scan_path(path)
        try:
            site = WordpressSite(
                    path=path,
                    structure_options=structure_options,
                    allow_io_errors=self.config.allow_io_errors
                )
        except WordpressException as error:
            if self.config.allow_io_errors:
                log.warning(
                        'Unable to scan site at ' + os.fsdecode(path)
                        + f': {error}'
                    )
                return
            else:
                raise
        log.debug('Located WordPress files at ' + os.fsdecode(site.core_path))
        version = site.get_version()
        log.debug(
                'WordPress Core Version: ' +
                version_to_str(version)
            )
        if scan_path is None:
            scan_path = path
        scanner.scan_core(version, scan_path)
        if check_extensions:
            self._scan_plugins(
                    site.get_all_plugins(self.config.allow_io_errors),
                    scanner,
                    scan_path,
                    True
                )
            self._scan_themes(
                    site.get_themes(self.config.allow_io_errors),
                    scanner,
                    scan_path,
                    True
                )

    def _get_vulnerability_label(self, count: int) -> str:
        if count == 1:
            return 'vulnerability'
        else:
            return 'vulnerabilities'

    def _output_summary(self, scanner: VulnerabilityScanner) -> None:
        unique_count = scanner.get_vulnerability_count()
        total_count = scanner.get_total_count()
        unique_label = self._get_vulnerability_label(unique_count)
        total_label = self._get_vulnerability_label(total_count)
        log.info(
                f'Found {unique_count} unique {unique_label} / {total_count} '
                f'total {total_label}'
            )

    def _validate_vulnerability_ids(
                self,
                identifiers: List[str],
                feed_variant: VulnerabilityFeedVariant) -> Set[str]:
        valid = set()
        for identifier in identifiers:
            if is_cve_id(identifier):
                if feed_variant is not VulnerabilityFeedVariant.PRODUCTION:
                    raise Exception(
                            'CVE IDs can only be used to filter '
                            'vulnerabilities with the production feed'
                        )
                valid.add(identifier)
            else:
                try:
                    uuid = UUID(identifier)
                    valid.add(str(uuid))
                except ValueError:
                    raise Exception(
                            f'Malformed vulnerability ID: {identifier}'
                        )
        return valid

    def _initialize_filter(
                self,
                feed_variant: VulnerabilityFeedVariant
            ) -> VulnerabilityFilter:
        excluded = self._validate_vulnerability_ids(
                self.config.exclude_vulnerability,
                feed_variant
            )
        included = self._validate_vulnerability_ids(
                self.config.include_vulnerability,
                feed_variant
            )
        return VulnerabilityFilter(
                excluded=excluded,
                included=included,
                informational=self.config.informational
            )

    def _scan_sites(
                self,
                path: bytes,
                scanner: VulnerabilityScanner,
                structure_options: WordpressStructureOptions = None
            ) -> None:
        log.info(
                'Searching for WordPress installations under '
                + os.fsdecode(path) + '...'
            )
        locator = WordpressLocator(
                path=path,
                allow_nested=self.config.allow_nested,
                allow_io_errors=self.config.allow_io_errors
            )
        site_found = False
        for core_path in locator.locate_core_paths():
            site_found = True
            log.info('Scanning site at ' + os.fsdecode(core_path) + '...')
            try:
                self._scan(
                        core_path,
                        scanner,
                        check_extensions=True,
                        structure_options=structure_options,
                        scan_path=path
                    )
            except AlreadyScannedException:
                log.warning(
                        'Site found at ' + os.fsdecode(core_path)
                        + ' has already been scanned'
                    )
        if not site_found:
            log.warning('No sites found under ' + os.fsdecode(path))

    def _requires_paths(self) -> bool:
        required = self.config.require_path
        return self.context.requires_input(required)

    def _check_required_paths(self) -> bool:
        if not self._requires_paths():
            return True
        return (len(self.config.trailing_arguments) +
                len(self.config.wordpress_path) +
                len(self.config.plugin_directory) +
                len(self.config.theme_directory)) > 0

    def _raise_path_error(self) -> None:
        raise VulnScanningConfigurationException(
                'At least one WordPress path must be specified'
            )

    def invoke(self) -> int:
        feed_variant = VulnerabilityFeedVariant.for_path(self.config.feed)
        report_manager = VulnScanReportManager(self.context, feed_variant)
        io_manager = report_manager.get_io_manager()
        if not io_manager.should_read_stdin() and \
                not self._check_required_paths():
            self._raise_path_error()
        if self.config.output_format == 'human' \
                and not self.context.allows_color:
            log.warning(
                    'The human output format requires a terminal with color '
                    'support to function properly. See --output-format for '
                    'other options.'
                )
        vulnerability_index = self._load_vulnerability_index(feed_variant)
        vulnerability_filter = self._initialize_filter(feed_variant)
        for invalid_id in vulnerability_filter.get_invalid_ids(
                    vulnerability_index
                ):
            log.warning(
                    f'Unrecognized vulnerability identifier: {invalid_id}, '
                    'expected a valid UUID or CVE ID'
                )
        scanner = VulnerabilityScanner(
                vulnerability_index,
                self._initialize_filter(feed_variant)
            )
        structure_options = WordpressStructureOptions(
                relative_content_paths=self.config.relative_content_path,
                relative_plugins_paths=self.config.relative_plugins_path,
                relative_mu_plugins_paths=self.config.relative_mu_plugins_path
            )
        with report_manager.open_output_file() as output_file:
            report = report_manager.initialize_report(output_file)
            scanner.register_result_callback(report.add_result)
            for path in self.config.trailing_arguments:
                self._scan_sites(
                        path,
                        scanner,
                        structure_options=structure_options
                    )
            if io_manager.should_read_stdin():
                reader = io_manager.get_input_reader()
                path_count = 0
                for path in reader.read_all_entries():
                    self._scan_sites(
                            path,
                            scanner,
                            structure_options=structure_options
                        )
                    path_count += 1
                if self._requires_paths() and path_count == 0:
                    self._raise_path_error()
            for path in self.config.wordpress_path:
                log.info(
                        'Scanning core installation at '
                        + os.fsdecode(path) + '...'
                    )
                try:
                    self._scan(
                            os.fsencode(path),
                            scanner,
                            structure_options=structure_options
                        )
                except AlreadyScannedException:
                    log.warning(
                            'Core installation at ' + os.fsdecode(path)
                            + ' has already been scanned'
                        )
            for path in self.config.plugin_directory:
                log.info(
                        'Scanning plugin directory at ' + os.fsdecode(path)
                        + '...'
                    )
                try:
                    self._scan_plugin_directory(os.fsencode(path), scanner)
                except AlreadyScannedException:
                    log.warning(
                            'Plugin directory at ' + os.fsdecode(path)
                            + ' has already been scanned'
                        )
            for path in self.config.theme_directory:
                log.info(
                        'Scanning theme directory at ' + os.fsdecode(path)
                        + '...'
                    )
                try:
                    self._scan_theme_directory(os.fsencode(path), scanner)
                except AlreadyScannedException:
                    log.warning(
                            'Theme directory at ' + os.fsdecode(path)
                            + ' has already been scanned'
                        )
            self._output_summary(scanner)
            report.scanner = scanner
            report.complete()
        return 0


factory = VulnScanSubcommand



from ctypes import c_char_p, c_void_p, c_int, c_ulong, c_ubyte, byref, \
        Structure, POINTER, CFUNCTYPE
from enum import IntEnum
from typing import Optional

from ..library import load_library, LibraryNotAvailableException
from ..encoding import bytes_to_str

from .pcre import PcreException, PcreLibraryNotAvailableException, \
    PcreOptions, \
    PCRE_DEFAULT_OPTIONS

try:
    pcre = load_library('pcre')
except LibraryNotAvailableException:
    raise PcreLibraryNotAvailableException('Failed to load libpcre')


_pcre_version = pcre.pcre_version
_pcre_version.argtypes = []
_pcre_version.restype = c_char_p
VERSION = bytes_to_str(_pcre_version())


class PcreError(IntEnum):
    NOMATCH = -1
    NULL = -2
    BADOPTION = -3
    BADMAGIC = -4
    UNKNOWN_OPCODE = -5
    NOMEMORY = -6
    NOSUBSTRING = -7
    MATCHLIMIT = -8
    CALLOUT = -9
    BADUTF = -10
    BASEUTF_OFFSET = -11
    PARTIAL = -12
    BADPARTIAL = -13
    INTERNAL = -14
    BADCOUNT = -15
    DFA_UITEM = -16
    DFA_COND = -17
    DFA_UMLIMIT = -18
    DFA_WSSIZE = -19
    DFA_RECURSE = -20
    RECURSION_LIMIT = -21
    NULLWSLIMIT = -22
    BADNEWLINE = -23
    BADOFFSET = -24
    SHORTUTF = -25
    RECURSELOOP = -26
    JIT_STACKLIMIT = -27
    BADMODE = -28
    BADENDIANNESS = -29
    DFA_BADRESTART = -30
    JIT_BADOPTION = -31
    BADLENGTH = -32
    UNSET = -33


_pcre_config = pcre.pcre_config
_pcre_config.argtypes = [c_int, c_void_p]
_pcre_config.restype = c_int

PCRE_CONFIG_JIT = 9


def _check_jit_support() -> bool:
    value = c_int(0)
    result = _pcre_config(PCRE_CONFIG_JIT, byref(value))
    if result == 0:
        return value.value == 1
    raise PcreException(f'Checking for JIT support failed, code {result}')


HAS_JIT_SUPPORT = _check_jit_support()


class _StructPcre(Structure):
    pass


_pcre_p = POINTER(_StructPcre)
_pcre_compile = pcre.pcre_compile
_pcre_compile.argtypes = [
        c_char_p,
        c_int,
        POINTER(c_char_p),
        POINTER(c_int),
        POINTER(c_ubyte)
    ]
_pcre_compile.restype = _pcre_p


class _StructPcreExtra(Structure):
    _fields_ = [
                ('flags', c_ulong),
                ('study_data', c_void_p),
                ('match_limit', c_ulong),
                ('callout_data', c_void_p),
                ('tables', POINTER(c_ubyte)),
                ('match_limit_recursion', c_ulong),
                ('mark', POINTER(c_char_p)),
                ('executable_jit', c_void_p)
            ]


_pcre_extra_p = POINTER(_StructPcreExtra)
_pcre_study = pcre.pcre_study
_pcre_study.argtypes = [_pcre_p, c_int, POINTER(c_char_p)]
_pcre_study.restype = _pcre_extra_p


_pcre_free_study = pcre.pcre_free_study
_pcre_free_study.argtypes = [_pcre_extra_p]
_pcre_free_study.restype = None


_pcre_exec = pcre.pcre_exec
_pcre_exec.argtypes = [
        _pcre_p,
        _pcre_extra_p,
        c_char_p,
        c_int,
        c_int,
        c_int,
        c_void_p,
        c_int
    ]
_pcre_exec.restype = c_int


_pcre_free_address = c_void_p.in_dll(pcre, 'pcre_free').value
_pcre_free_prototype = CFUNCTYPE(None, c_void_p)
_pcre_free = _pcre_free_prototype(_pcre_free_address)


if HAS_JIT_SUPPORT:
    class _StructPcreJitStack(Structure):
        pass

    _pcre_jit_stack_p = POINTER(_StructPcreJitStack)

    _pcre_jit_stack_alloc = pcre.pcre_jit_stack_alloc
    _pcre_jit_stack_alloc.argtypes = [c_int, c_int]
    _pcre_jit_stack_alloc.restype = _pcre_jit_stack_p

    _pcre_jit_stack_free = pcre.pcre_jit_stack_free
    _pcre_jit_stack_free.argtypes = [_pcre_jit_stack_p]
    _pcre_jit_stack_free.restype = None

    _pcre_jit_exec = pcre.pcre_jit_exec
    _pcre_jit_exec.argtypes = [
            _pcre_p,
            _pcre_extra_p,
            c_char_p,
            c_int,
            c_int,
            c_int,
            c_void_p,
            c_int,
            _pcre_jit_stack_p
        ]
    _pcre_jit_exec.restype = c_int


PCRE_EXTRA_MATCH_LIMIT = 0x0012
PCRE_EXTRA_MATCH_LIMIT_RECURSION = 0x0010
PCRE_STUDY_JIT_COMPILE = 0x0001
PCRE_STUDY_EXTRA_NEEDED = 0x0008


PCRE_JIT_STACK_MIN_SIZE = 32 * 1024
PCRE_JIT_STACK_MAX_SIZE = 64 * 1024


class PcreJitStack:

    def __init__(
                self,
                min_size: int = PCRE_JIT_STACK_MIN_SIZE,
                max_size: int = PCRE_JIT_STACK_MAX_SIZE
            ):
        self.min_size = min_size
        self.max_size = max_size
        self._jit_stack = None

    def allocate(self) -> None:
        if HAS_JIT_SUPPORT:
            self._jit_stack = _pcre_jit_stack_alloc(
                    self.min_size,
                    self.max_size
                )

    def _allocate_if_necessary(self) -> None:
        if self._jit_stack is None:
            self.allocate()

    def free(self) -> None:
        if HAS_JIT_SUPPORT and self._jit_stack is not None:
            _pcre_jit_stack_free(self._jit_stack)

    def __enter__(self):
        self._allocate_if_necessary()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.free()

    def _get_jit_stack(self):
        self._allocate_if_necessary()
        return self._jit_stack


class PcreMatch:

    def __init__(self, matched_string: bytes):
        self.matched_string = matched_string


class PcrePattern:

    def __init__(
                self,
                pattern: str,
                options: PcreOptions = PCRE_DEFAULT_OPTIONS
            ):
        self.pattern = pattern
        self.options = options
        self._compile()

    def _compile(self) -> None:
        pattern_cstr = c_char_p(self.pattern.encode('utf8'))
        error_message = c_char_p(None)
        error_offset = c_int(-1)
        self.compiled = _pcre_compile(
                pattern_cstr,
                self.options._get_compilation_options(),
                byref(error_message),
                byref(error_offset),
                None
            )
        if not self.compiled:
            offset = error_offset.value
            message = error_message.value.decode('utf8')
            raise PcreException(
                    f'Pattern compilation failed at offset {offset}: {message}'
                )
        study_options = c_int(PCRE_STUDY_JIT_COMPILE | PCRE_STUDY_EXTRA_NEEDED)
        self.extra = _pcre_study(
                self.compiled,
                study_options,
                byref(error_message)
            )
        self.extra.flags = c_ulong(
                PCRE_EXTRA_MATCH_LIMIT | PCRE_EXTRA_MATCH_LIMIT_RECURSION
            )
        self.extra.match_limit = c_ulong(self.options.match_limit)
        self.extra.match_limit_recursion = \
            c_ulong(self.options.match_limit_recursion)

    def match(
                self,
                subject: bytes,
                jit_stack: PcreJitStack = None
            ) -> Optional[PcreMatch]:
        if jit_stack is None and HAS_JIT_SUPPORT:
            jit_stack = PcreJitStack()
            temporary_jit_stack = True
        else:
            temporary_jit_stack = False
        subject_cstr = c_char_p(subject)
        subject_length = c_int(len(subject))
        start_offset = c_int(0)
        options = c_int(0)
        ovector = c_int * 3
        ovector = ovector(0, 0, 0)
        ovecsize = c_int(3)
        if HAS_JIT_SUPPORT:
            result = _pcre_jit_exec(
                    self.compiled,
                    self.extra,
                    subject_cstr,
                    subject_length,
                    start_offset,
                    options,
                    byref(ovector),
                    ovecsize,
                    jit_stack._get_jit_stack()
                )
        else:
            result = _pcre_exec(
                    self.compiled,
                    self.extra,
                    subject_cstr,
                    subject_length,
                    start_offset,
                    options,
                    byref(ovector),
                    ovecsize,
                )
        if temporary_jit_stack:
            jit_stack.free()
        if result < 0:
            try:
                error = PcreError(result)
                if error is PcreError.NOMATCH:
                    return None
                else:
                    raise PcreException(
                            'Matching failed with error: '
                            f'{error.name}({error.value})'
                        )
            except ValueError:
                raise PcreException(
                        f'Matching failed with unknown error: {result}'
                    )
        else:
            matched_string = subject[ovector[0]:ovector[1]]
            return PcreMatch(matched_string)

    def _free(self) -> None:
        if self.extra is not None and _pcre_free_study is not None:
            _pcre_free_study(self.extra)
            self.extra = None
        if self.compiled is not None and _pcre_free is not None:
            _pcre_free(self.compiled)
            self.compiled = None

    def __del__(self) -> None:
        self._free()

    def __getstate__(self) -> dict:
        return {
                'pattern': self.pattern,
                'options': self.options
            }

    def __setstate__(self, state) -> None:
        self.pattern = state['pattern']
        self.options = state['options']
        self._compile()
